# -*- coding: utf-8 -*-
# dirsnapshot.py: Directory snapshots and comparison.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
:module: watchdog.utils.dirsnapshot
:synopsis: Directory snapshots and comparison.
:author: Gora Khargosh <gora.khargosh@gmail.com>

.. ADMONITION:: Where are the moved events? They "disappeared"

        This implementation does not take partition boundaries
        into consideration. It will only work when the directory
        tree is entirely on the same file system. More specifically,
        any part of the code that depends on inode numbers can
        break if partition boundaries are crossed. In these cases,
        the snapshot diff will represent file/directory movement as
        created and deleted events.

        Windows does not have any concept of ``inodes``, which prevents
        this snapshotter from determining file or directory renames/movement
        on it. The snapshotter does not try to handle this on Windows.
        File or directory movement will show up as creation and deletion
        events.

        Please do not use this on a virtual file system mapped to
        a network share.

Classes
-------
.. autoclass:: DirectorySnapshot
   :members:
   :show-inheritance:

.. autoclass:: DirectorySnapshotDiff
   :members:
   :show-inheritance:

"""

import os
import sys
import stat

from pathtools.path import walk as path_walk, absolute_path

if not sys.version < (2, 6, 0):
    from watchdog.utils.bricks import OrderedSet as set

class DirectorySnapshotDiff(object):
    """
    Compares two directory snapshots and creates an object that represents
    the difference between the two snapshots.

    :param ref_dirsnap:
        The reference directory snapshot.
    :type ref_dirsnap:
        :class:`DirectorySnapshot`
    :param dirsnap:
        The directory snapshot which will be compared
        with the reference snapshot.
    :type dirsnap:
        :class:`DirectorySnapshot`
    """
    def __init__(self, ref_dirsnap, dirsnap):
        """
        """
        self._files_deleted = list()
        self._files_modified = list()
        self._files_created = list()
        self._files_moved = list()

        self._dirs_modified = list()
        self._dirs_moved = list()
        self._dirs_deleted = list()
        self._dirs_created = list()

        # Detect all the modifications.
        for path, stat_info in dirsnap.stat_snapshot.items():
            if path in ref_dirsnap.stat_snapshot:
                ref_stat_info = ref_dirsnap.stat_info(path)
                if stat_info.st_ino == ref_stat_info.st_ino and stat_info.st_mtime != ref_stat_info.st_mtime:
                    if stat.S_ISDIR(stat_info.st_mode):
                        self._dirs_modified.append(path)
                    else:
                        self._files_modified.append(path)

        paths_deleted = ref_dirsnap.paths - dirsnap.paths
        paths_created = dirsnap.paths - ref_dirsnap.paths

        # Detect all the moves/renames.
        # Doesn't work on Windows, so exlude on Windows.
        if not sys.platform.startswith('win'):
            for created_path in paths_created.copy():
                created_stat_info = dirsnap.stat_info(created_path)
                for deleted_path in paths_deleted.copy():
                    deleted_stat_info = ref_dirsnap.stat_info(deleted_path)
                    if created_stat_info.st_ino == deleted_stat_info.st_ino:
                        paths_deleted.remove(deleted_path)
                        paths_created.remove(created_path)
                        if stat.S_ISDIR(created_stat_info.st_mode):
                            self._dirs_moved.append((deleted_path, created_path))
                        else:
                            self._files_moved.append((deleted_path, created_path))

        # Now that we have renames out of the way, enlist the deleted and
        # created files/directories.
        for path in paths_deleted:
            stat_info = ref_dirsnap.stat_info(path)
            if stat.S_ISDIR(stat_info.st_mode):
                self._dirs_deleted.append(path)
            else:
                self._files_deleted.append(path)

        for path in paths_created:
            stat_info = dirsnap.stat_info(path)
            if stat.S_ISDIR(stat_info.st_mode):
                self._dirs_created.append(path)
            else:
                self._files_created.append(path)


    @property
    def files_created(self):
        """List of files that were created."""
        return self._files_created

    @property
    def files_deleted(self):
        """List of files that were deleted."""
        return self._files_deleted

    @property
    def files_modified(self):
        """List of files that were modified."""
        return self._files_modified

    @property
    def files_moved(self):
        """
        List of files that were moved.

        Each event is a two-tuple the first item of which is the path
        that has been renamed to the second item in the tuple.
        """
        return self._files_moved

    @property
    def dirs_modified(self):
        """
        List of directories that were modified.
        """
        return self._dirs_modified

    @property
    def dirs_moved(self):
        """
        List of directories that were moved.

        Each event is a two-tuple the first item of which is the path
        that has been renamed to the second item in the tuple.
        """
        return self._dirs_moved

    @property
    def dirs_deleted(self):
        """
        List of directories that were deleted.
        """
        return self._dirs_deleted

    @property
    def dirs_created(self):
        """
        List of directories that were created.
        """
        return self._dirs_created


class DirectorySnapshot(object):
    """
    A snapshot of stat information of files in a directory.

    :param path:
        The directory path for which a snapshot should be taken.
    :type path:
        ``str``
    :param recursive:
        ``True`` if the entired directory tree should be included in the
        snapshot; ``False`` otherwise.
    :type recursive:
        ``bool``
    :param walker_callback:
        A function with the signature ``walker_callback(path, stat_info)``
        which will be called for every entry in the directory tree.
    """
    def __init__(self,
                 path,
                 recursive=True,
                 walker_callback=(lambda p, s: None),
                 _copying=False):
        self._path = absolute_path(path)
        self._stat_snapshot = {}
        self._inode_to_path = {}
        self.is_recursive = recursive

        if not _copying:
            stat_info = os.stat(self._path)
            self._stat_snapshot[self._path] = stat_info
            self._inode_to_path[stat_info.st_ino] = self._path
            walker_callback(self._path, stat_info)

            for root, directories, files in path_walk(self._path, recursive):
                for directory_name in directories:
                    try:
                        directory_path = os.path.join(root, directory_name)
                        stat_info = os.stat(directory_path)
                        self._stat_snapshot[directory_path] = stat_info
                        self._inode_to_path[stat_info.st_ino] = directory_path
                        walker_callback(directory_path, stat_info)
                    except OSError:
                        continue

                for file_name in files:
                    try:
                        file_path = os.path.join(root, file_name)
                        stat_info = os.stat(file_path)
                        self._stat_snapshot[file_path] = stat_info
                        self._inode_to_path[stat_info.st_ino] = file_path
                        walker_callback(file_path, stat_info)
                    except OSError:
                        continue


    def __sub__(self, previous_dirsnap):
        """Allow subtracting a DirectorySnapshot object instance from
        another.

        :returns:
            A :class:`DirectorySnapshotDiff` object.
        """
        return DirectorySnapshotDiff(previous_dirsnap, self)

    #def __add__(self, new_dirsnap):
    #    self._stat_snapshot.update(new_dirsnap._stat_snapshot)

    def copy(self, from_pathname=None):
        snapshot = DirectorySnapshot(path=from_pathname,
                                     recursive=self.is_recursive,
                                     _copying=True)
        for pathname, stat_info in self._stat_snapshot.items():
            if pathname.starts_with(from_pathname):
                snapshot._stat_snapshot[pathname] = stat_info
                snapshot._inode_to_path[stat_info.st_ino] = pathname
        return snapshot


    @property
    def stat_snapshot(self):
        """
        Returns a dictionary of stat information with file paths being keys.
        """
        return self._stat_snapshot


    def stat_info(self, path):
        """
        Returns a stat information object for the specified path from
        the snapshot.

        :param path:
            The path for which stat information should be obtained
            from a snapshot.
        """
        return self._stat_snapshot[path]


    def path_for_inode(self, inode):
        """
        Determines the path that an inode represents in a snapshot.

        :param inode:
            inode number.
        """
        return self._inode_to_path[inode]


    def stat_info_for_inode(self, inode):
        """
        Determines stat information for a given inode.

        :param inode:
            inode number.
        """
        return self.stat_info(self.path_for_inode(inode))


    @property
    def paths(self):
        """
        List of file/directory paths in the snapshot.
        """
        return set(self._stat_snapshot)


    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return str(self._stat_snapshot)




# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
:module: watchdog.observers
:synopsis: Observer that picks a native implementation if available.
:author: Gora Khargosh <gora.khargosh@gmail.com>


Classes
=======
.. autoclass:: Observer
   :members:
   :show-inheritance:
   :inherited-members:

"""

from watchdog.observers.api import BaseObserver, DEFAULT_OBSERVER_TIMEOUT

# Ensure FSEvents is checked *before* kqueue here. Mac OS X supports
# both FSEvents and kqueue, and FSEvents is the preferred way of monitoring
# file system events on this OS.
try: # pragma: no cover
    from watchdog.observers.inotify import InotifyObserver as _Observer
except ImportError: # pragma: no cover
    try: # pragma: no cover
        from watchdog.observers.fsevents import FSEventsObserver_BUGGY_TODO as _Observer
    except ImportError: # pragma: no cover
        try: # pragma: no cover
            from watchdog.observers.kqueue import KqueueObserver as _Observer
        except ImportError: # pragma: no cover
            try: # pragma: no cover
                from watchdog.observers.read_directory_changes_async import WindowsApiAsyncObserver as _Observer
            except ImportError: # pragma: no cover
                try: # pragma: no cover
                    from watchdog.observers.read_directory_changes import WindowsApiObserver as _Observer
                except ImportError: # pragma: no cover
                    from watchdog.observers.polling import PollingObserver as _Observer


class Observer(_Observer):
    """
    Observer thread that schedules watching directories and dispatches
    calls to event handlers.
    """



# -*- coding: utf-8 -*-
# winapi.py: Windows API-Python interface (removes dependency on pywin32)
#
# Copyright (C) 2007 Thomas Heller <theller@ctypes.org>
# Copyright (C) 2010 Will McGugan <will@willmcgugan.com>
# Copyright (C) 2010 Ryan Kelly <ryan@rfk.id.au>
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and / or other materials provided with the distribution.
# * Neither the name of the organization nor the names of its contributors may
#   be used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Portions of this code were taken from pyfilesystem, which uses the above
# new BSD license.

from __future__ import with_statement
from watchdog.utils import platform

if platform.is_windows():
    import ctypes.wintypes
    import struct

    try:
        LPVOID = ctypes.wintypes.LPVOID
    except AttributeError:
        # LPVOID wasn't defined in Py2.5, guess it was introduced in Py2.6
        LPVOID = ctypes.c_void_p

    # Invalid handle value.
    INVALID_HANDLE_VALUE = 0xFFFFFFFF     # -1

    # File notification contants.
    FILE_NOTIFY_CHANGE_FILE_NAME = 0x01
    FILE_NOTIFY_CHANGE_DIR_NAME = 0x02
    FILE_NOTIFY_CHANGE_ATTRIBUTES = 0x04
    FILE_NOTIFY_CHANGE_SIZE = 0x08
    FILE_NOTIFY_CHANGE_LAST_WRITE = 0x010
    FILE_NOTIFY_CHANGE_LAST_ACCESS = 0x020
    FILE_NOTIFY_CHANGE_CREATION = 0x040
    FILE_NOTIFY_CHANGE_SECURITY = 0x0100

    FILE_FLAG_BACKUP_SEMANTICS = 0x02000000
    FILE_FLAG_OVERLAPPED = 0x40000000
    FILE_LIST_DIRECTORY = 0x01
    FILE_SHARE_READ = 0x01
    FILE_SHARE_WRITE = 0x02
    FILE_SHARE_DELETE = 0x04
    OPEN_EXISTING = 3

    # File action constants.
    FILE_ACTION_CREATED = 1
    FILE_ACTION_DELETED = 2
    FILE_ACTION_MODIFIED = 3
    FILE_ACTION_RENAMED_OLD_NAME = 4
    FILE_ACTION_RENAMED_NEW_NAME = 5
    FILE_ACTION_OVERFLOW = 0xFFFF

    # Aliases
    FILE_ACTION_ADDED = FILE_ACTION_CREATED
    FILE_ACTION_REMOVED = FILE_ACTION_DELETED

    THREAD_TERMINATE = 0x0001

    # IO waiting constants.
    WAIT_ABANDONED = 0x00000080
    WAIT_IO_COMPLETION = 0x000000C0
    WAIT_OBJECT_0 = 0x00000000
    WAIT_TIMEOUT = 0x00000102


    class OVERLAPPED(ctypes.Structure):
        _fields_ = [('Internal', LPVOID),
                    ('InternalHigh', LPVOID),
                    ('Offset', ctypes.wintypes.DWORD),
                    ('OffsetHigh', ctypes.wintypes.DWORD),
                    ('Pointer', LPVOID),
                    ('hEvent', ctypes.wintypes.HANDLE),
                   ]

    def _errcheck_bool(value, func, args):
        if not value:
            raise ctypes.WinError()
        return args

    def _errcheck_handle(value, func, args):
        if not value:
            raise ctypes.WinError()
        if value == INVALID_HANDLE_VALUE:
            raise ctypes.WinError()
        return args

    def _errcheck_dword(value, func, args):
        if value == 0xFFFFFFFF:
            raise ctypes.WinError()
        return args

    try:
        ReadDirectoryChangesW = ctypes.windll.kernel32.ReadDirectoryChangesW
    except AttributeError:
        raise ImportError("ReadDirectoryChangesW is not available")
    ReadDirectoryChangesW.restype = ctypes.wintypes.BOOL
    ReadDirectoryChangesW.errcheck = _errcheck_bool
    ReadDirectoryChangesW.argtypes = (
        ctypes.wintypes.HANDLE, # hDirectory
        LPVOID, # lpBuffer
        ctypes.wintypes.DWORD, # nBufferLength
        ctypes.wintypes.BOOL, # bWatchSubtree
        ctypes.wintypes.DWORD, # dwNotifyFilter
        ctypes.POINTER(ctypes.wintypes.DWORD), # lpBytesReturned
        ctypes.POINTER(OVERLAPPED), # lpOverlapped
        LPVOID #FileIOCompletionRoutine # lpCompletionRoutine
    )

    CreateFileW = ctypes.windll.kernel32.CreateFileW
    CreateFileW.restype = ctypes.wintypes.HANDLE
    CreateFileW.errcheck = _errcheck_handle
    CreateFileW.argtypes = (
        ctypes.wintypes.LPCWSTR, # lpFileName
        ctypes.wintypes.DWORD, # dwDesiredAccess
        ctypes.wintypes.DWORD, # dwShareMode
        LPVOID, # lpSecurityAttributes
        ctypes.wintypes.DWORD, # dwCreationDisposition
        ctypes.wintypes.DWORD, # dwFlagsAndAttributes
        ctypes.wintypes.HANDLE # hTemplateFile
    )

    CloseHandle = ctypes.windll.kernel32.CloseHandle
    CloseHandle.restype = ctypes.wintypes.BOOL
    CloseHandle.argtypes = (
        ctypes.wintypes.HANDLE, # hObject
    )

    CreateEvent = ctypes.windll.kernel32.CreateEventW
    CreateEvent.restype = ctypes.wintypes.HANDLE
    CreateEvent.errcheck = _errcheck_handle
    CreateEvent.argtypes = (
        LPVOID, # lpEventAttributes
        ctypes.wintypes.BOOL, # bManualReset
        ctypes.wintypes.BOOL, # bInitialState
        ctypes.wintypes.LPCWSTR, #lpName
    )

    SetEvent = ctypes.windll.kernel32.SetEvent
    SetEvent.restype = ctypes.wintypes.BOOL
    SetEvent.errcheck = _errcheck_bool
    SetEvent.argtypes = (
        ctypes.wintypes.HANDLE, # hEvent
    )

    WaitForSingleObjectEx = ctypes.windll.kernel32.WaitForSingleObjectEx
    WaitForSingleObjectEx.restype = ctypes.wintypes.DWORD
    WaitForSingleObjectEx.errcheck = _errcheck_dword
    WaitForSingleObjectEx.argtypes = (
        ctypes.wintypes.HANDLE, # hObject
        ctypes.wintypes.DWORD, # dwMilliseconds
        ctypes.wintypes.BOOL, # bAlertable
    )

    CreateIoCompletionPort = ctypes.windll.kernel32.CreateIoCompletionPort
    CreateIoCompletionPort.restype = ctypes.wintypes.HANDLE
    CreateIoCompletionPort.errcheck = _errcheck_handle
    CreateIoCompletionPort.argtypes = (
        ctypes.wintypes.HANDLE, # FileHandle
        ctypes.wintypes.HANDLE, # ExistingCompletionPort
        LPVOID, # CompletionKey
        ctypes.wintypes.DWORD, # NumberOfConcurrentThreads
    )

    GetQueuedCompletionStatus = ctypes.windll.kernel32.GetQueuedCompletionStatus
    GetQueuedCompletionStatus.restype = ctypes.wintypes.BOOL
    GetQueuedCompletionStatus.errcheck = _errcheck_bool
    GetQueuedCompletionStatus.argtypes = (
        ctypes.wintypes.HANDLE, # CompletionPort
        LPVOID, # lpNumberOfBytesTransferred
        LPVOID, # lpCompletionKey
        ctypes.POINTER(OVERLAPPED), # lpOverlapped
        ctypes.wintypes.DWORD, # dwMilliseconds
    )

    PostQueuedCompletionStatus = ctypes.windll.kernel32.PostQueuedCompletionStatus
    PostQueuedCompletionStatus.restype = ctypes.wintypes.BOOL
    PostQueuedCompletionStatus.errcheck = _errcheck_bool
    PostQueuedCompletionStatus.argtypes = (
        ctypes.wintypes.HANDLE, # CompletionPort
        ctypes.wintypes.DWORD, # lpNumberOfBytesTransferred
        ctypes.wintypes.DWORD, # lpCompletionKey
        ctypes.POINTER(OVERLAPPED), # lpOverlapped
    )

    class FILE_NOTIFY_INFORMATION(ctypes.Structure):
        _fields_ = [("NextEntryOffset", ctypes.wintypes.DWORD),
                    ("Action", ctypes.wintypes.DWORD),
                    ("FileNameLength", ctypes.wintypes.DWORD),
                    #("FileName", (ctypes.wintypes.WCHAR * 1))]
                    ("FileName", (ctypes.c_char * 1))]
    LPFNI = ctypes.POINTER(FILE_NOTIFY_INFORMATION)

    def get_FILE_NOTIFY_INFORMATION(readBuffer, nBytes):
        results = []
        while nBytes > 0:
            fni = ctypes.cast(readBuffer, LPFNI)[0]
            ptr = ctypes.addressof(fni) + FILE_NOTIFY_INFORMATION.FileName.offset
            #filename = ctypes.wstring_at(ptr, fni.FileNameLength)
            filename = ctypes.string_at(ptr, fni.FileNameLength)
            results.append((fni.Action, filename.decode('utf-16')))
            numToSkip = fni.NextEntryOffset
            if numToSkip <= 0:
                break
            readBuffer = readBuffer[numToSkip:]
            nBytes -= numToSkip # numToSkip is long. nBytes should be long too.
        return results

    def get_FILE_NOTIFY_INFORMATION_alt(event_buffer, nBytes):
        """Extract the information out of a FILE_NOTIFY_INFORMATION structure."""
        pos = 0
        event_buffer = event_buffer[:nBytes]
        while pos < len(event_buffer):
            jump, action, namelen = struct.unpack("iii", event_buffer[pos:pos + 12])
            # TODO: this may return a shortname or a longname, with no way
            # to tell which.  Normalise them somehow?
            name = event_buffer[pos + 12:pos + 12 + namelen].decode("utf-16")
            yield (name, action)
            if not jump:
                break
            pos += jump



# -*- coding: utf-8 -*-
# api.py: Observer, event emitter, and event queue API.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
:module: watchdog.observers.api
:synopsis: Classes useful to observer implementers.
:author: Gora Khargosh <gora.khargosh@gmail.com>

Immutables
----------
.. autoclass:: ObservedWatch
   :members:
   :show-inheritance:


Collections
-----------
.. autoclass:: EventQueue
   :members:
   :show-inheritance:

Classes
-------
.. autoclass:: EventEmitter
   :members:
   :show-inheritance:

.. autoclass:: EventDispatcher
   :members:
   :show-inheritance:

.. autoclass:: BaseObserver
   :members:
   :show-inheritance:
"""

from __future__ import with_statement
import threading
try:
    import queue # IGNORE:F0401
except ImportError:
    import Queue as queue # IGNORE:F0401

from pathtools.path import absolute_path
from watchdog.utils import DaemonThread
from brownie.datastructures.queues import SetQueue

DEFAULT_EMITTER_TIMEOUT = 1    # in seconds.
DEFAULT_OBSERVER_TIMEOUT = 1   # in seconds.


# Collection classes
class EventQueue(SetQueue):
    """Thread-safe event queue based on a thread-safe ordered-set queue
    to ensure duplicate :class:`FileSystemEvent` objects are prevented from
    adding themselves to the queue to avoid dispatching multiple event handling
    calls when multiple identical events are produced quicker than an observer
    can consume them.
    """

class ObservedWatch(object):
    """An scheduled watch.

    :param path:
        Path string.
    :param recursive:
        ``True`` if watch is recursive; ``False`` otherwise.
    """
    def __init__(self, path, recursive):
        self._path = absolute_path(path)
        self._is_recursive = recursive

    @property
    def path(self):
        """The path that this watch monitors."""
        return self._path

    @property
    def is_recursive(self):
        """Determines whether subdirectories are watched for the path."""
        return self._is_recursive

    @property
    def key(self):
        return (self.path, self.is_recursive)

    def __eq__(self, watch):
        return self.key == watch.key

    def __ne__(self, watch):
        return self.key != watch.key

    def __hash__(self):
        return hash(self.key)

    def __repr__(self):
        return "<ObservedWatch: path=%s, is_recursive=%s>" % (self.path, self.is_recursive)



# Observer classes
class EventEmitter(DaemonThread):
    """
    Producer daemon thread base class subclassed by event emitters
    that generate events and populate a queue with them.

    :param event_queue:
        The event queue to populate with generated events.
    :type event_queue:
        :class:`watchdog.events.EventQueue`
    :param watch:
        The watch to observe and produce events for.
    :type watch:
        :class:`ObservedWatch`
    :param timeout:
        Timeout (in seconds) between successive attempts at reading events.
    :type timeout:
        ``float``
    """
    def __init__(self, event_queue, watch, timeout=DEFAULT_EMITTER_TIMEOUT):
        DaemonThread.__init__(self)
        self._event_queue = event_queue
        self._watch = watch
        self._timeout = timeout

    @property
    def timeout(self):
        """
        Blocking timeout for reading events.
        """
        return self._timeout

    @property
    def watch(self):
        """
        The watch associated with this emitter.
        """
        return self._watch

    def queue_event(self, event):
        """
        Queues a single event.

        :param event:
            Event to be queued.
        :type event:
            An instance of :class:`watchdog.events.FileSystemEvent`
            or a subclass.
        """
        self._event_queue.put((event, self.watch))

    def queue_events(self, timeout):
        """Override this method to populate the event queue with events
        per interval period.

        :param timeout:
            Timeout (in seconds) between successive attempts at
            reading events.
        :type timeout:
            ``float``
        """

    def on_thread_exit(self):
        """
        Override this method for cleaning up immediately before the daemon
        thread stops completely.
        """

    def run(self):
        try:
            while self.should_keep_running():
                self.queue_events(self.timeout)
        finally:
            self.on_thread_exit()


class EventDispatcher(DaemonThread):
    """
    Consumer daemon thread base class subclassed by event observer threads
    that dispatch events from an event queue to appropriate event handlers.

    :param timeout:
        Event queue blocking timeout (in seconds).
    :type timeout:
        ``float``
    """
    def __init__(self, timeout=DEFAULT_OBSERVER_TIMEOUT):
        DaemonThread.__init__(self)
        self._event_queue = EventQueue()
        self._timeout = timeout

    @property
    def timeout(self):
        """Event queue block timeout."""
        return self._timeout

    @property
    def event_queue(self):
        """The event queue which is populated with file system events
        by emitters and from which events are dispatched by a dispatcher
        thread."""
        return self._event_queue

    def dispatch_event(self, event, watch):
        """Override this method to dispatch an individual event.

        :param event:
            Event to be dispatched.
        :type event:
            An instance of a subclass of
            :class:`watchdog.events.FileSystemEvent`
        :param watch:
            The watch to dispatch for.
        :type watch:
            An instance of :class:`ObservedWatch` or a subclass of
            :class:`ObservedWatch`
        """

    def _dispatch_events(self, event_queue, timeout):
        """Consumes events from an event queue. Blocks on the queue for the
        specified timeout before raising :class:`queue.Empty`.

        :param event_queue:
            Event queue to populate with one set of events.
        :type event_queue:
            :class:`EventQueue`
        :param timeout:
            Interval period (in seconds) to wait before timing out on the
            event queue.
        :type timeout:
            ``float``
        :raises:
            :class:`queue.Empty`
        """
        event, watch = event_queue.get(block=True, timeout=timeout)
        self.dispatch_event(event, watch)
        event_queue.task_done()

    def on_thread_exit(self):
        """Override this method for cleaning up immediately before the daemon
        thread stops completely."""

    def run(self):
        try:
            while self.should_keep_running():
                try:
                    self._dispatch_events(self.event_queue, self.timeout)
                except queue.Empty:
                    continue
        finally:
            self.on_thread_exit()


class BaseObserver(EventDispatcher):
    """Base observer."""
    def __init__(self, emitter_class, timeout=DEFAULT_OBSERVER_TIMEOUT):
        EventDispatcher.__init__(self, timeout)
        self._emitter_class = emitter_class
        self._lock = threading.Lock()
        self._watches = set()
        self._handlers = dict()
        self._emitters = set()
        self._emitter_for_watch = dict()


    def _add_emitter(self, emitter):
        self._emitter_for_watch[emitter.watch] = emitter
        self._emitters.add(emitter)

    def _remove_emitter(self, emitter):
        del self._emitter_for_watch[emitter.watch]
        self._emitters.remove(emitter)
        emitter.stop()

    def _get_emitter_for_watch(self, watch):
        return self._emitter_for_watch[watch]

    def _clear_emitters(self):
        for emitter in self._emitters:
            emitter.stop()
        self._emitters.clear()
        self._emitter_for_watch.clear()

    def _add_handler_for_watch(self, event_handler, watch):
        try:
            self._handlers[watch].add(event_handler)
        except KeyError:
            self._handlers[watch] = set([event_handler])

    def _get_handlers_for_watch(self, watch):
        return self._handlers[watch]

    def _remove_handlers_for_watch(self, watch):
        del self._handlers[watch]

    def _remove_handler_for_watch(self, handler, watch):
        handlers = self._get_handlers_for_watch(watch)
        handlers.remove(handler)


    def schedule(self, event_handler, path, recursive=False):
        """
        Schedules watching a path and calls appropriate methods specified
        in the given event handler in response to file system events.

        :param event_handler:
            An event handler instance that has appropriate event handling
            methods which will be called by the observer in response to
            file system events.
        :type event_handler:
            :class:`watchdog.events.FileSystemEventHandler` or a subclass
        :param path:
            Directory path that will be monitored.
        :type path:
            ``str``
        :param recursive:
            ``True`` if events will be emitted for sub-directories
            traversed recursively; ``False`` otherwise.
        :type recursive:
            ``bool``
        :return:
            An :class:`ObservedWatch` object instance representing
            a watch.
        """
        with self._lock:
            watch = ObservedWatch(path, recursive)
            self._add_handler_for_watch(event_handler, watch)
            try:
                # If we have an emitter for this watch already, we don't create a
                # new emitter. Instead we add the handler to the event
                # object.
                emitter = self._get_emitter_for_watch(watch)
            except KeyError:
                # Create a new emitter and start it.
                emitter = self._emitter_class(event_queue=self.event_queue,
                                              watch=watch,
                                              timeout=self.timeout)
                self._add_emitter(emitter)
                emitter.start()
            self._watches.add(watch)
        return watch

    def add_handler_for_watch(self, event_handler, watch):
        """Adds a handler for the given watch.

        :param event_handler:
            An event handler instance that has appropriate event handling
            methods which will be called by the observer in response to
            file system events.
        :type event_handler:
            :class:`watchdog.events.FileSystemEventHandler` or a subclass
        :param watch:
            The watch to add a handler for.
        :type watch:
            An instance of :class:`ObservedWatch` or a subclass of
            :class:`ObservedWatch`
        """
        with self._lock:
            self._add_handler_for_watch(event_handler, watch)

    def remove_handler_for_watch(self, event_handler, watch):
        """Removes a handler for the given watch.

        :param event_handler:
            An event handler instance that has appropriate event handling
            methods which will be called by the observer in response to
            file system events.
        :type event_handler:
            :class:`watchdog.events.FileSystemEventHandler` or a subclass
        :param watch:
            The watch to remove a handler for.
        :type watch:
            An instance of :class:`ObservedWatch` or a subclass of
            :class:`ObservedWatch`
        """
        with self._lock:
            self._remove_handler_for_watch(event_handler, watch)

    def unschedule(self, watch):
        """Unschedules a watch.

        :param watch:
            The watch to unschedule.
        :type watch:
            An instance of :class:`ObservedWatch` or a subclass of
            :class:`ObservedWatch`
        """
        with self._lock:
            try:
                emitter = self._get_emitter_for_watch(watch)
                self._remove_handlers_for_watch(watch)
                self._remove_emitter(emitter)
                self._watches.remove(watch)
            except KeyError:
                raise

    def unschedule_all(self):
        """Unschedules all watches and detaches all associated event
        handlers."""
        with self._lock:
            self._handlers.clear()
            self._clear_emitters()
            self._watches.clear()

    def on_thread_exit(self):
        self.unschedule_all()

    def dispatch_event(self, event, watch):
        with self._lock:
            for handler in self._get_handlers_for_watch(watch):
                handler.dispatch(event)



# -*- coding: utf-8 -*-
# platform.py: platform determination.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys

PLATFORM_WINDOWS = 'windows'
PLATFORM_LINUX = 'linux'
PLATFORM_BSD = 'bsd'
PLATFORM_DARWIN = 'darwin'
PLATFORM_UNKNOWN = 'unknown'

def get_platform_name():
    if sys.platform.startswith("win"):
        return PLATFORM_WINDOWS
    elif sys.platform.startswith('darwin'):
        return PLATFORM_DARWIN
    elif sys.platform.startswith('linux'):
        return PLATFORM_LINUX
    elif sys.platform.startswith('bsd'):
        return PLATFORM_BSD
    else:
        return PLATFORM_UNKNOWN

__platform__ = get_platform_name()

def is_linux():
    return __platform__ == PLATFORM_LINUX

def is_bsd():
    return __platform__ == PLATFORM_BSD

def is_darwin():
    return __platform__ == PLATFORM_DARWIN

def is_windows():
    return __platform__ == PLATFORM_WINDOWS




# -*- coding: utf-8 -*-
# fsevents.py: FSEvents-based event emitter for Mac OS X.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
:module: watchdog.observers.fsevents
:synopsis: FSEvents based emitter implementation.
:author: Gora Khargosh <gora.khargosh@gmail.com>
:platforms: Mac OS X
"""

from __future__ import with_statement
from watchdog.utils import platform

if platform.is_darwin():
    import threading
    import unicodedata
    import os.path
    import _watchdog_fsevents as _fsevents

    from pathtools.path import absolute_path
    from watchdog.events import\
        FileDeletedEvent,\
        FileModifiedEvent,\
        FileCreatedEvent,\
        FileMovedEvent,\
        DirDeletedEvent,\
        DirModifiedEvent,\
        DirCreatedEvent,\
        DirMovedEvent,\
        DirMovedEvent
    from watchdog.utils.dirsnapshot import DirectorySnapshot
    from watchdog.observers.api import\
        BaseObserver,\
        EventEmitter,\
        DEFAULT_EMITTER_TIMEOUT,\
        DEFAULT_OBSERVER_TIMEOUT


    class FSEventsEmitter(EventEmitter):
        """
        Mac OS X FSEvents Emitter class.

        :param event_queue:
            The event queue to fill with events.
        :param watch:
            A watch object representing the directory to monitor.
        :type watch:
            :class:`watchdog.observers.api.ObservedWatch`
        :param timeout:
            Read events blocking timeout (in seconds).
        :type timeout:
            ``float``
        """

        def __init__(self, event_queue, watch, timeout=DEFAULT_EMITTER_TIMEOUT):
            EventEmitter.__init__(self, event_queue, watch, timeout)
            self._lock = threading.Lock()
            self.snapshot = DirectorySnapshot(watch.path, watch.is_recursive)

        def on_thread_exit(self):
            _fsevents.remove_watch(self.watch)
            _fsevents.stop(self)

        def queue_events(self, timeout):
            with self._lock:
                if not self.watch.is_recursive\
                and self.watch.path not in pathnames:
                    return
                new_snapshot = DirectorySnapshot(self.watch.path,
                                                 self.watch.is_recursive)
                events = new_snapshot - self.snapshot
                self.snapshot = new_snapshot

                # Files.
                for src_path in events.files_deleted:
                    self.queue_event(FileDeletedEvent(src_path))
                for src_path in events.files_modified:
                    self.queue_event(FileModifiedEvent(src_path))
                for src_path in events.files_created:
                    self.queue_event(FileCreatedEvent(src_path))
                for src_path, dest_path in events.files_moved:
                    self.queue_event(FileMovedEvent(src_path, dest_path))

                # Directories.
                for src_path in events.dirs_deleted:
                    self.queue_event(DirDeletedEvent(src_path))
                for src_path in events.dirs_modified:
                    self.queue_event(DirModifiedEvent(src_path))
                for src_path in events.dirs_created:
                    self.queue_event(DirCreatedEvent(src_path))
                for src_path, dest_path in events.dirs_moved:
                    self.queue_event(DirMovedEvent(src_path, dest_path))


        def run(self):
            try:
                def callback(pathnames, flags, emitter=self):
                    emitter.queue_events(emitter.timeout)

                #for pathname, flag in zip(pathnames, flags):
                #if emitter.watch.is_recursive: # and pathname != emitter.watch.path:
                #    new_sub_snapshot = DirectorySnapshot(pathname, True)
                #    old_sub_snapshot = self.snapshot.copy(pathname)
                #    diff = new_sub_snapshot - old_sub_snapshot
                #    self.snapshot += new_subsnapshot
                #else:
                #    new_snapshot = DirectorySnapshot(emitter.watch.path, False)
                #    diff = new_snapshot - emitter.snapshot
                #    emitter.snapshot = new_snapshot


                # INFO: FSEvents reports directory notifications recursively
                # by default, so we do not need to add subdirectory paths.
                #pathnames = set([self.watch.path])
                #if self.watch.is_recursive:
                #    for root, directory_names, _ in os.walk(self.watch.path):
                #        for directory_name in directory_names:
                #            full_path = absolute_path(
                #                            os.path.join(root, directory_name))
                #            pathnames.add(full_path)

                pathnames = [self.watch.path]
                _fsevents.add_watch(self,
                                    self.watch,
                                    callback,
                                    pathnames)
                _fsevents.read_events(self)
            finally:
                self.on_thread_exit()


    class FSEventsObserver(BaseObserver):
        def __init__(self, timeout=DEFAULT_OBSERVER_TIMEOUT):
            BaseObserver.__init__(self, emitter_class=FSEventsEmitter,
                                  timeout=timeout)

        def schedule(self, event_handler, path, recursive=False):
            # Fix for issue #26: Trace/BPT error when given a unicode path
            # string. https://github.com/gorakhargosh/watchdog/issues#issue/26
            if isinstance(path, unicode):
                #path = unicode(path, 'utf-8')
                path = unicodedata.normalize('NFC', path).encode('utf-8')
            BaseObserver.schedule(self, event_handler, path, recursive)



# -*- coding: utf-8 -*-
# winapi_common.py: Common routines and constants for the Win32 API used.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from watchdog.utils import platform

if platform.is_windows():
    import ctypes

    from watchdog.observers.winapi import \
        FILE_FLAG_BACKUP_SEMANTICS, \
        FILE_FLAG_OVERLAPPED, \
        FILE_SHARE_READ, \
        FILE_SHARE_WRITE, \
        FILE_SHARE_DELETE, \
        FILE_NOTIFY_CHANGE_FILE_NAME, \
        FILE_NOTIFY_CHANGE_DIR_NAME, \
        FILE_NOTIFY_CHANGE_ATTRIBUTES, \
        FILE_NOTIFY_CHANGE_SIZE, \
        FILE_NOTIFY_CHANGE_LAST_WRITE, \
        FILE_NOTIFY_CHANGE_SECURITY, \
        FILE_NOTIFY_CHANGE_LAST_ACCESS, \
        FILE_NOTIFY_CHANGE_CREATION, \
        FILE_ACTION_CREATED, \
        FILE_ACTION_DELETED, \
        FILE_ACTION_MODIFIED, \
        FILE_LIST_DIRECTORY, \
        OPEN_EXISTING, \
        INVALID_HANDLE_VALUE, \
        CreateFileW, \
        ReadDirectoryChangesW, \
        CreateIoCompletionPort, \
        CloseHandle
    from watchdog.events import \
        DirDeletedEvent, \
        DirCreatedEvent, \
        DirModifiedEvent, \
        FileDeletedEvent, \
        FileCreatedEvent, \
        FileModifiedEvent

    # We don't need to recalculate these flags every time a call is made to
    # the win32 API functions.
    WATCHDOG_FILE_FLAGS = FILE_FLAG_BACKUP_SEMANTICS
    WATCHDOG_FILE_FLAGS_ASYNC = FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OVERLAPPED
    WATCHDOG_FILE_SHARE_FLAGS = reduce(lambda x, y: x | y, [
        FILE_SHARE_READ,
        FILE_SHARE_WRITE,
        FILE_SHARE_DELETE,
    ])
    WATCHDOG_FILE_NOTIFY_FLAGS = reduce(lambda x, y: x | y, [
        FILE_NOTIFY_CHANGE_FILE_NAME,
        FILE_NOTIFY_CHANGE_DIR_NAME,
        FILE_NOTIFY_CHANGE_ATTRIBUTES,
        FILE_NOTIFY_CHANGE_SIZE,
        FILE_NOTIFY_CHANGE_LAST_WRITE,
        FILE_NOTIFY_CHANGE_SECURITY,
        FILE_NOTIFY_CHANGE_LAST_ACCESS,
        FILE_NOTIFY_CHANGE_CREATION,
    ])


    # HACK:
    WATCHDOG_TRAVERSE_MOVED_DIR_DELAY = 1   # seconds

    BUFFER_SIZE = 2048

    # Moved event is handled explicitly in the emitter thread.
    DIR_ACTION_EVENT_MAP = {
        FILE_ACTION_CREATED: DirCreatedEvent,
        FILE_ACTION_DELETED: DirDeletedEvent,
        FILE_ACTION_MODIFIED: DirModifiedEvent,
    }
    FILE_ACTION_EVENT_MAP = {
        FILE_ACTION_CREATED: FileCreatedEvent,
        FILE_ACTION_DELETED: FileDeletedEvent,
        FILE_ACTION_MODIFIED: FileModifiedEvent,
    }


    def get_directory_handle(path, file_flags):
        """Returns a Windows handle to the specified directory path."""
        handle = CreateFileW(path,
                             FILE_LIST_DIRECTORY,
                             WATCHDOG_FILE_SHARE_FLAGS,
                             None,
                             OPEN_EXISTING,
                             file_flags,
                             None)
        return handle


    def close_directory_handle(handle):
        CloseHandle(handle)


    def read_directory_changes(handle, event_buffer, recursive):
        """Read changes to the directory using the specified directory handle.

        http://timgolden.me.uk/pywin32-docs/win32file__ReadDirectoryChangesW_meth.html
        """
        nbytes = ctypes.wintypes.DWORD()
        ReadDirectoryChangesW(handle,
                              ctypes.byref(event_buffer),
                              len(event_buffer),
                              recursive,
                              WATCHDOG_FILE_NOTIFY_FLAGS,
                              ctypes.byref(nbytes),
                              None,
                              None)
        # get_FILE_NOTIFY_INFORMATION expects nBytes to be long.
        return event_buffer.raw, long(nbytes.value)


    def create_io_completion_port():
        """
        http://timgolden.me.uk/pywin32-docs/win32file__CreateIoCompletionPort_meth.html
        """
        return CreateIoCompletionPort(INVALID_HANDLE_VALUE, None, 0, 0)



# -*- coding: utf-8 -*-
# watchdog.tricks.__init__.py: tricks framework.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import subprocess

from watchdog.utils import echo, has_attribute
from watchdog.events import PatternMatchingEventHandler


class Trick(PatternMatchingEventHandler):
    """Your tricks should subclass this class."""

    @classmethod
    def generate_yaml(cls):
        context = dict(module_name=cls.__module__,
                       klass_name=cls.__name__)
        template_yaml = """- %(module_name)s.%(klass_name)s:
  args:
  - argument1
  - argument2
  kwargs:
    patterns:
    - "*.py"
    - "*.js"
    ignore_patterns:
    - "version.py"
    ignore_directories: false
"""
        return template_yaml % context


class LoggerTrick(Trick):
    """A simple trick that does only logs events."""
    def on_any_event(self, event):
        pass

    @echo.echo
    def on_modified(self, event):
        pass

    @echo.echo
    def on_deleted(self, event):
        pass

    @echo.echo
    def on_created(self, event):
        pass

    @echo.echo
    def on_moved(self, event):
        pass


class ShellCommandTrick(Trick):
    """Execeutes shell commands in response to matched events."""
    def __init__(self, shell_command=None, patterns=None, ignore_patterns=None, ignore_directories=False):
        super(ShellCommandTrick, self).__init__(patterns, ignore_patterns, ignore_directories)
        self.shell_command = shell_command

    def on_any_event(self, event):
        from string import Template
        if event.is_directory:
            object_type = 'directory'
        else:
            object_type = 'file'

        context = {
            'watch_src_path': event.src_path,
            'watch_dest_path': '',
            'watch_event_type': event.event_type,
            'watch_object': object_type,
            }

        if self.shell_command is None:
            if has_attribute(event, 'dest_path'):
                context.update({'dest_path': event.dest_path})
                command = 'echo "${watch_event_type} ${watch_object} from ${watch_src_path} to ${watch_dest_path}"'
            else:
                command = 'echo "${watch_event_type} ${watch_object} ${watch_src_path}"'
        else:
            if has_attribute(event, 'dest_path'):
                context.update({'watch_dest_path': event.dest_path})
            command = self.shell_command

        command = Template(command).safe_substitute(**context)
        subprocess.Popen(command, shell=True)



# -*- coding: utf-8 -*-
# polling.py: Generic polling emitter implementation.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
:module: watchdog.observers.polling
:synopsis: Polling emitter implementation.
:author: Gora Khargosh <gora.khargosh@gmail.com>

Classes
-------
.. autoclass:: PollingEmitter
   :members:
   :show-inheritance:
"""


from __future__ import with_statement

import time
import threading

from watchdog.utils.dirsnapshot import DirectorySnapshot, DirectorySnapshotDiff
from watchdog.observers.api import \
    EventEmitter, \
    BaseObserver, \
    DEFAULT_OBSERVER_TIMEOUT, \
    DEFAULT_EMITTER_TIMEOUT
from watchdog.events import \
    DirMovedEvent, \
    DirDeletedEvent, \
    DirCreatedEvent, \
    DirModifiedEvent, \
    FileMovedEvent, \
    FileDeletedEvent, \
    FileCreatedEvent, \
    FileModifiedEvent


class PollingEmitter(EventEmitter):
    """
    Platform-independent emitter that polls a directory to detect file
    system changes.
    """
    def __init__(self, event_queue, watch, timeout=DEFAULT_EMITTER_TIMEOUT):
        EventEmitter.__init__(self, event_queue, watch, timeout)
        self._snapshot = DirectorySnapshot(watch.path, watch.is_recursive)
        self._lock = threading.Lock()

    def on_thread_exit(self):
        self._snapshot = None


    def queue_events(self, timeout):
        with self._lock:
            # We don't want to hit the disk continuously.
            # timeout behaves like an interval for polling emitters.
            time.sleep(timeout)

            # Get event diff between fresh snapshot and previous snapshot.
            # Update snapshot.
            new_snapshot = DirectorySnapshot(self.watch.path, self.watch.is_recursive)
            events = DirectorySnapshotDiff(self._snapshot, new_snapshot)
            self._snapshot = new_snapshot

            # Files.
            for src_path in events.files_deleted:
                self.queue_event(FileDeletedEvent(src_path))
            for src_path in events.files_modified:
                self.queue_event(FileModifiedEvent(src_path))
            for src_path in events.files_created:
                self.queue_event(FileCreatedEvent(src_path))
            for src_path, dest_path in events.files_moved:
                self.queue_event(FileMovedEvent(src_path, dest_path))

            # Directories.
            for src_path in events.dirs_deleted:
                self.queue_event(DirDeletedEvent(src_path))
            for src_path in events.dirs_modified:
                self.queue_event(DirModifiedEvent(src_path))
            for src_path in events.dirs_created:
                self.queue_event(DirCreatedEvent(src_path))
            for src_path, dest_path in events.dirs_moved:
                self.queue_event(DirMovedEvent(src_path, dest_path))



class PollingObserver(BaseObserver):
    """
    Observer thread that schedules watching directories and dispatches
    calls to event handlers.
    """
    def __init__(self, timeout=DEFAULT_OBSERVER_TIMEOUT):
        BaseObserver.__init__(self, emitter_class=PollingEmitter, timeout=timeout)





# -*- coding: utf-8 -*-
# utils.py: Utility functions.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
:module: watchdog.utils
:synopsis: Utility classes and functions.
:author: Gora Khargosh <gora.khargosh@gmail.com>

Functions
---------

.. autofunction:: has_attribute

.. autofunction:: load_class

.. autofunction:: load_module

.. autofunction:: read_text_file

Classes
-------
.. autoclass:: DaemonThread
   :members:
   :show-inheritance:
   :inherited-members:

"""

import os
import os.path
import sys
import threading
import ctypes.util

from fnmatch import fnmatch


def ctypes_find_library(name, default):
    """Finds a dynamic library."""
    module_path = None
    try:
        module_path = ctypes.util.find_library(name)
    except (OSError, IOError):
        module_path = default
    return module_path


def has_attribute(ob, attribute):
    """
    :func:`hasattr` swallows exceptions. :func:`has_attribute` tests a Python object for the
    presence of an attribute.

    :param ob:
        object to inspect
    :param attribute:
        ``str`` for the name of the attribute.
    """
    return getattr(ob, attribute, None) is not None


class DaemonThread(threading.Thread):
    """
    Daemon thread convenience class, sets a few properties and makes
    writing daemon threads a little easier.
    """
    def __init__(self):
        threading.Thread.__init__(self)
        if has_attribute(self, 'daemon'):
            self.daemon = True
        else:
            self.setDaemon(True)
        self._stopped_event = threading.Event()

        if not has_attribute(self._stopped_event.is_set, 'is_set'):
            self._stopped_event.is_set = self._stopped_event.isSet

    @property
    def stopped_event(self):
        return self._stopped_event

    def should_stop(self):
        """Determines whether the daemon thread should stop."""
        return self._stopped_event.is_set()

    def should_keep_running(self):
        """Determines whether the daemon thread should continue running."""
        return not self._stopped_event.is_set()

    def on_thread_told_to_stop(self):
        """Override this method instead of :meth:`stop()`.
        :meth:`stop()` calls this method.

        Note that this method is called immediately after the daemon thread
        is signaled to halt.
        """
        pass

    def stop(self):
        """Signals the daemon thread to stop."""
        self._stopped_event.set()
        self.on_thread_told_to_stop()


if not has_attribute(DaemonThread, 'is_alive'):
    DaemonThread.is_alive = DaemonThread.isAlive


def load_module(module_name):
    """Imports a module given its name and returns a handle to it."""
    try:
        __import__(module_name)
    except ImportError:
        raise ImportError('No module named %s' % module_name)
    return sys.modules[module_name]


def load_class(dotted_path):
    """Loads and returns a class definition provided a dotted path
    specification the last part of the dotted path is the class name
    and there is at least one module name preceding the class name.

    Notes:
    You will need to ensure that the module you are trying to load
    exists in the Python path.

    Examples:
    - module.name.ClassName    # Provided module.name is in the Python path.
    - module.ClassName         # Provided module is in the Python path.

    What won't work:
    - ClassName
    - modle.name.ClassName     # Typo in module name.
    - module.name.ClasNam      # Typo in classname.
    """
    dotted_path_split = dotted_path.split('.')
    if len(dotted_path_split) > 1:
        klass_name = dotted_path_split[-1]
        module_name = '.'.join(dotted_path_split[:-1])

        module = load_module(module_name)
        if has_attribute(module, klass_name):
            klass = getattr(module, klass_name)
            return klass
            # Finally create and return an instance of the class
            #return klass(*args, **kwargs)
        else:
            raise AttributeError('Module %s does not have class attribute %s' % (module_name, klass_name))
    else:
        raise ValueError('Dotted module path %s must contain a module name and a classname' % dotted_path)


def read_text_file(file_path, mode='rb'):
    """
    Returns the contents of a file after opening it in read-only mode.

    :param file_path:
        Path to the file to be read from.
    :param mode:
        Mode string.
    """
    return open(file_path, mode).read()




# -*- coding: utf-8 -*-
# events.py: File system events and event handlers.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
:module: watchdog.events
:synopsis: File system events and event handlers.
:author: Gora Khargosh <gora.khargosh@gmail.com>

Event Classes
-------------
.. autoclass:: FileSystemEvent
   :members:
   :show-inheritance:
   :inherited-members:

.. autoclass:: FileSystemMovedEvent
   :members:
   :show-inheritance:

.. autoclass:: FileMovedEvent
   :members:
   :show-inheritance:

.. autoclass:: DirMovedEvent
   :members:
   :show-inheritance:

.. autoclass:: FileModifiedEvent
   :members:
   :show-inheritance:

.. autoclass:: DirModifiedEvent
   :members:
   :show-inheritance:

.. autoclass:: FileCreatedEvent
   :members:
   :show-inheritance:

.. autoclass:: DirCreatedEvent
   :members:
   :show-inheritance:

.. autoclass:: FileDeletedEvent
   :members:
   :show-inheritance:

.. autoclass:: DirDeletedEvent
   :members:
   :show-inheritance:


Event Handler Classes
---------------------
.. autoclass:: FileSystemEventHandler
   :members:
   :show-inheritance:

.. autoclass:: PatternMatchingEventHandler
   :members:
   :show-inheritance:

.. autoclass:: LoggingEventHandler
   :members:
   :show-inheritance:

"""

import os.path
import logging

from pathtools.path import absolute_path
from pathtools.patterns import match_any_paths
from watchdog.utils import has_attribute


EVENT_TYPE_MOVED = 'moved'
EVENT_TYPE_DELETED = 'deleted'
EVENT_TYPE_CREATED = 'created'
EVENT_TYPE_MODIFIED = 'modified'

class FileSystemEvent(object):
    """
    Immutable type that represents a file system event that is triggered
    when a change occurs on the monitored file system.

    All FileSystemEvent objects are required to be immutable and hence
    can be used as keys in dictionaries or be added to sets.
    """

    def __init__(self, event_type, src_path, is_directory=False):
        self._src_path = src_path
        self._is_directory = is_directory
        self._event_type = event_type

    @property
    def is_directory(self):
        """True if event was emitted for a directory; False otherwise."""
        return self._is_directory

    @property
    def src_path(self):
        """Source path of the file system object that triggered this event."""
        return self._src_path

    @property
    def event_type(self):
        """The type of the event as a string."""
        return self._event_type

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%(class_name)s: event_type=%(event_type)s, \
src_path=%(src_path)s, is_directory=%(is_directory)s>" %\
               dict(class_name=self.__class__.__name__,
                    event_type=self.event_type,
                    src_path=self.src_path,
                    is_directory=self.is_directory)

    # Used for comparison of events.
    @property
    def key(self):
        return (self.event_type,
                self.src_path,
                self.is_directory)

    def __eq__(self, event):
        return self.key == event.key

    def __ne__(self, event):
        return self.key != event.key

    def __hash__(self):
        return hash(self.key)


class FileSystemMovedEvent(FileSystemEvent):
    """
    File system event representing any kind of file system movement.
    """

    def __init__(self, src_path, dest_path, is_directory):
        super(FileSystemMovedEvent, self).__init__(event_type=EVENT_TYPE_MOVED,
                                                   src_path=src_path,
                                                   is_directory=is_directory)

        self._dest_path = dest_path

    @property
    def dest_path(self):
        """The destination path of the move event."""
        return self._dest_path

    # Used for hashing this as an immutable object.
    def _key(self):
        return (self.event_type,
                self.src_path,
                self.dest_path,
                self.is_directory)

    def __repr__(self):
        return "<%(class_name)s: src_path=%(src_path)s, dest_path=%(dest_path)s, \
is_directory=%(is_directory)s>" %\
               dict(class_name=self.__class__.__name__,
                    src_path=self.src_path,
                    dest_path=self.dest_path,
                    is_directory=self.is_directory)


    # File events.

class FileDeletedEvent(FileSystemEvent):
    """File system event representing file deletion on the file system."""

    def __init__(self, src_path):
        super(FileDeletedEvent, self).__init__(event_type=EVENT_TYPE_DELETED,
                                               src_path=src_path)

    def __repr__(self):
        return "<%(class_name)s: src_path=%(src_path)s>" %\
               dict(class_name=self.__class__.__name__,
                    src_path=self.src_path)

class FileModifiedEvent(FileSystemEvent):
    """File system event representing file modification on the file system."""

    def __init__(self, src_path):
        super(FileModifiedEvent, self).__init__(event_type=EVENT_TYPE_MODIFIED,
                                                src_path=src_path)

    def __repr__(self):
        return "<%(class_name)s: src_path=%(src_path)s>" %\
               dict(class_name=self.__class__.__name__,
                    src_path=self.src_path)


class FileCreatedEvent(FileSystemEvent):
    """File system event representing file creation on the file system."""

    def __init__(self, src_path):
        super(FileCreatedEvent, self).__init__(event_type=EVENT_TYPE_CREATED,
                                               src_path=src_path)

    def __repr__(self):
        return "<%(class_name)s: src_path=%(src_path)s>" %\
               dict(class_name=self.__class__.__name__,
                    src_path=self.src_path)


class FileMovedEvent(FileSystemMovedEvent):
    """File system event representing file movement on the file system."""

    def __init__(self, src_path, dest_path):
        super(FileMovedEvent, self).__init__(src_path=src_path,
                                             dest_path=dest_path,
                                             is_directory=False)

    def __repr__(self):
        return "<%(class_name)s: src_path=%(src_path)s, \
dest_path=%(dest_path)s>" %\
               dict(class_name=self.__class__.__name__,
                    src_path=self.src_path,
                    dest_path=self.dest_path)


    # Directory events.

class DirDeletedEvent(FileSystemEvent):
    """File system event representing directory deletion on the file system."""

    def __init__(self, src_path):
        super(DirDeletedEvent, self).__init__(event_type=EVENT_TYPE_DELETED,
                                              src_path=src_path,
                                              is_directory=True)

    def __repr__(self):
        return "<%(class_name)s: src_path=%(src_path)s>" %\
               dict(class_name=self.__class__.__name__,
                    src_path=self.src_path)


class DirModifiedEvent(FileSystemEvent):
    """
    File system event representing directory modification on the file system.
    """

    def __init__(self, src_path):
        super(DirModifiedEvent, self).__init__(event_type=EVENT_TYPE_MODIFIED,
                                               src_path=src_path,
                                               is_directory=True)

    def __repr__(self):
        return "<%(class_name)s: src_path=%(src_path)s>" %\
               dict(class_name=self.__class__.__name__,
                    src_path=self.src_path)


class DirCreatedEvent(FileSystemEvent):
    """File system event representing directory creation on the file system."""

    def __init__(self, src_path):
        super(DirCreatedEvent, self).__init__(event_type=EVENT_TYPE_CREATED,
                                              src_path=src_path,
                                              is_directory=True)

    def __repr__(self):
        return "<%(class_name)s: src_path=%(src_path)s>" %\
               dict(class_name=self.__class__.__name__,
                    src_path=self.src_path)


class DirMovedEvent(FileSystemMovedEvent):
    """File system event representing directory movement on the file system."""

    def __init__(self, src_path, dest_path):
        super(DirMovedEvent, self).__init__(src_path=src_path,
                                            dest_path=dest_path,
                                            is_directory=True)

    def __repr__(self):
        return "<%(class_name)s: src_path=%(src_path)s, \
dest_path=%(dest_path)s>" %\
               dict(class_name=self.__class__.__name__,
                    src_path=self.src_path,
                    dest_path=self.dest_path)


    def sub_moved_events(self, _walker=os.walk):
        """Generates moved events for file sytem objects within the
        moved directory.

        :param _walker:
            Walker used to walk directory trees :func:`os.walk` style. Sanity
            tests use this parameter to inject a mock walker that behaves like
            :func:`os.walk`.
        :returns:
            iterable of event objects of type :class:`FileMovedEvent` and
            :class:`DirMovedEvent`.
        """
        return list(
                _generate_sub_moved_events_for(self.src_path, self.dest_path,
                                               _walker=_walker))


class FileSystemEventHandler(object):
    """Base file system event handler that you can override methods from.
    """

    def dispatch(self, event):
        """Dispatches events to the appropriate methods.

        :param event:
            The event object representing the file system event.
        :type event:
            :class:`FileSystemEvent`
        """
        self.on_any_event(event)
        _method_map = {
            EVENT_TYPE_MODIFIED: self.on_modified,
            EVENT_TYPE_MOVED: self.on_moved,
            EVENT_TYPE_CREATED: self.on_created,
            EVENT_TYPE_DELETED: self.on_deleted,
        }
        event_type = event.event_type
        _method_map[event_type](event)

    def on_any_event(self, event):
        """Catch-all event handler.

        :param event:
            The event object representing the file system event.
        :type event:
            :class:`FileSystemEvent`
        """

    def on_moved(self, event):
        """Called when a file or a directory is moved or renamed.

        :param event:
            Event representing file/directory movement.
        :type event:
            :class:`DirMovedEvent` or :class:`FileMovedEvent`
        """

    def on_created(self, event):
        """Called when a file or directory is created.

        :param event:
            Event representing file/directory creation.
        :type event:
            :class:`DirCreatedEvent` or :class:`FileCreatedEvent`
        """


    def on_deleted(self, event):
        """Called when a file or directory is deleted.

        :param event:
            Event representing file/directory deletion.
        :type event:
            :class:`DirDeletedEvent` or :class:`FileDeletedEvent`
        """

    def on_modified(self, event):
        """Called when a file or directory is modified.

        :param event:
            Event representing file/directory modification.
        :type event:
            :class:`DirModifiedEvent` or :class:`FileModifiedEvent`
        """


class PatternMatchingEventHandler(FileSystemEventHandler):
    """
    Matches given patterns with file paths associated with occurring events.
    """

    def __init__(self, patterns=None, ignore_patterns=None,
                 ignore_directories=False, case_sensitive=False):
        super(PatternMatchingEventHandler, self).__init__()

        self._patterns = patterns
        self._ignore_patterns = ignore_patterns
        self._ignore_directories = ignore_directories
        self._case_sensitive = case_sensitive

    @property
    def patterns(self):
        """
        (Read-only)
        Patterns to allow matching event paths.
        """
        return self._patterns

    @property
    def ignore_patterns(self):
        """
        (Read-only)
        Patterns to ignore matching event paths.
        """
        return self._ignore_patterns

    @property
    def ignore_directories(self):
        """
        (Read-only)
        ``True`` if directories should be ignored; ``False`` otherwise.
        """
        return self._ignore_directories

    @property
    def case_sensitive(self):
        """
        (Read-only)
        ``True`` if path names should be matched sensitive to case; ``False``
        otherwise.
        """
        return self._case_sensitive

    def dispatch(self, event):
        """Dispatches events to the appropriate methods.

        :param event:
            The event object representing the file system event.
        :type event:
            :class:`FileSystemEvent`
        """
        if self.ignore_directories and event.is_directory:
            return

        if has_attribute(event, 'dest_path'):
            paths = [event.src_path, event.dest_path]
        else:
            paths = [event.src_path]

        if match_any_paths(paths,
                           included_patterns=self.patterns,
                           excluded_patterns=self.ignore_patterns,
                           case_sensitive=self.case_sensitive):
            self.on_any_event(event)
            _method_map = {
                EVENT_TYPE_MODIFIED: self.on_modified,
                EVENT_TYPE_MOVED: self.on_moved,
                EVENT_TYPE_CREATED: self.on_created,
                EVENT_TYPE_DELETED: self.on_deleted,
            }
            event_type = event.event_type
            _method_map[event_type](event)


class LoggingEventHandler(FileSystemEventHandler):
    """Logs all the events captured."""

    def on_moved(self, event):
        super(LoggingEventHandler, self).on_moved(event)

        what = 'directory' if event.is_directory else 'file'
        logging.info("Moved %s: from %s to %s", what, event.src_path,
                     event.dest_path)

    def on_created(self, event):
        super(LoggingEventHandler, self).on_created(event)

        what = 'directory' if event.is_directory else 'file'
        logging.info("Created %s: %s", what, event.src_path)

    def on_deleted(self, event):
        super(LoggingEventHandler, self).on_deleted(event)

        what = 'directory' if event.is_directory else 'file'
        logging.info("Deleted %s: %s", what, event.src_path)

    def on_modified(self, event):
        super(LoggingEventHandler, self).on_modified(event)

        what = 'directory' if event.is_directory else 'file'
        logging.info("Modified %s: %s", what, event.src_path)


class LoggingFileSystemEventHandler(LoggingEventHandler):
    """For backwards-compatibility. Please use :class:`LoggingEventHandler` instead."""


def _generate_sub_moved_events_for(src_dir_path, dest_dir_path,
                                   _walker=os.walk):
    """Generates an event list of :class:`DirMovedEvent` and :class:`FileMovedEvent`
    objects for all the files and directories within the given moved directory
    that were moved along with the directory.

    :param src_dir_path:
        The source path of the moved directory.
    :param dest_dir_path:
        The destination path of the moved directory.
    :param _walker:
        Walker used to walk directory trees :func:`os.walk` style. Sanity tests
        use this parameter to inject a mock walker that behaves like
        :func:`os.walk`.
    :returns:
        An iterable of file system events of type :class:`DirMovedEvent` and
        :class:`FileMovedEvent`.
    """
    src_dir_path = absolute_path(src_dir_path)
    dest_dir_path = absolute_path(dest_dir_path)
    for root, directories, filenames in _walker(dest_dir_path):
        for directory in directories:
            full_path = os.path.join(root, directory)
            renamed_path = full_path.replace(dest_dir_path, src_dir_path)
            yield DirMovedEvent(renamed_path, full_path)
        for filename in filenames:
            full_path = os.path.join(root, filename)
            renamed_path = full_path.replace(dest_dir_path, src_dir_path)
            yield FileMovedEvent(renamed_path, full_path)




# -*- coding: utf-8 -*-
# bricks.py: utility collections.
#
# Copyright (C) 2009, 2010 Raymond Hettinger <python@rcn.com>
# Copyright (C) 2010 Lukáš Lalinský <lalinsky@gmail.com>
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


"""
Utility collections or "bricks".

:module: watchdog.utils.bricks
:author: Gora Khargosh <gora.khargosh@gmail.com>
:author: Lukáš Lalinský <lalinsky@gmail.com>
:author: Raymond Hettinger <python@rcn.com>

Classes
=======
.. autoclass:: OrderedSetQueue
   :members:
   :show-inheritance:
   :inherited-members:

.. autoclass:: OrderedSet

"""

import sys
import collections
try:
    import queue
except ImportError:
    import Queue as queue

class OrderedSetQueue(queue.Queue):
    """Thread-safe implementation of an ordered set queue.

    Disallows adding a duplicate item while maintaining the
    order of items in the queue. The implementation leverages
    locking already implemented in the base class
    redefining only the primitives. Since the internal queue
    is not replaced, the order is maintained. The set is used
    merely to check for the existence of an item.

    Queued items must be immutable and hashable so that they can be used
    as dictionary keys. You must implement **only read-only properties** and
    the :meth:`Item.__hash__()`, :meth:`Item.__eq__()`, and
    :meth:`Item.__ne__()` methods for items to be hashable.

    An example implementation follows::

        class Item(object):
            def __init__(self, a, b):
                self._a = a
                self._b = b

            @property
            def a(self):
                return self._a

            @property
            def b(self):
                return self._b

            def _key(self):
                return (self._a, self._b)

            def __eq__(self, item):
                return self._key() == item._key()

            def __ne__(self, item):
                return self._key() != item._key()

            def __hash__(self):
                return hash(self._key())

    :author: Lukáš Lalinský <lalinsky@gmail.com>
    :url: http://stackoverflow.com/questions/1581895/how-check-if-a-task-is-already-in-python-queue
    """
    def _init(self, maxsize):
        queue.Queue._init(self, maxsize)
        self._set_of_items = set()

    def _put(self, item):
        if item not in self._set_of_items:
            queue.Queue._put(self, item)
            self._set_of_items.add(item)

    def _get(self):
        item = queue.Queue._get(self)
        self._set_of_items.remove(item)
        return item



if not sys.version < (2, 6, 0):
    KEY, PREV, NEXT = range(3)

    class OrderedSet(collections.MutableSet):
        """
        Implementation based on a doubly-linked link and an internal dictionary.
        This design gives :class:`OrderedSet` the same big-Oh running times as
        regular sets including O(1) adds, removes, and lookups as well as
        O(n) iteration.

        .. ADMONITION:: Implementation notes

                Runs on Python 2.6 or later (and runs on Python 3.0 or later
                without any modifications).

        :author: Raymond Hettinger <python@rcn.com>
        :url: http://code.activestate.com/recipes/576694/
        """
        def __init__(self, iterable=None):
            self.end = end = []
            end += [None, end, end]         # sentinel node for doubly linked list
            self.map = {}                   # key --> [key, prev, next]
            if iterable is not None:
                self |= iterable

        def __len__(self):
            return len(self.map)

        def __contains__(self, key):
            return key in self.map

        def add(self, key):
            if key not in self.map:
                end = self.end
                curr = end[PREV]
                curr[NEXT] = end[PREV] = self.map[key] = [key, curr, end]

        def discard(self, key):
            if key in self.map:
                key, prev, _next = self.map.pop(key)
                prev[NEXT] = _next
                _next[PREV] = prev

        def __iter__(self):
            end = self.end
            curr = end[NEXT]
            while curr is not end:
                yield curr[KEY]
                curr = curr[NEXT]

        def __reversed__(self):
            end = self.end
            curr = end[PREV]
            while curr is not end:
                yield curr[KEY]
                curr = curr[PREV]

        def pop(self, last=True):
            if not self:
                raise KeyError('set is empty')
            key = next(reversed(self)) if last else next(iter(self))
            self.discard(key)
            return key

        def __repr__(self):
            if not self:
                return '%s()' % (self.__class__.__name__,)
            return '%s(%r)' % (self.__class__.__name__, list(self))

        def __eq__(self, other):
            if isinstance(other, OrderedSet):
                return len(self) == len(other) and list(self) == list(other)
            return set(self) == set(other)

        def __del__(self):
            self.clear()                    # remove circular references




# -*- coding: utf-8 -*-
# inotify.py: inotify-based event emitter for Linux 2.6.13+.
#
# Copyright (C) 2010 Sebastien Martini <seb@dbzteam.org>
# Copyright (C) 2010 Luke McCarthy <luke@iogopro.co.uk>
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
# Copyright (C) 2010 Tim Cuthbertson <tim+github@gfxmonk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


"""
:module: watchdog.observers.inotify
:synopsis: ``inotify(7)`` based emitter implementation.
:author: Sebastien Martini <seb@dbzteam.org>
:author: Luke McCarthy <luke@iogopro.co.uk>
:author: Gora Khargosh <gora.khargosh@gmail.com>
:author: Tim Cuthbertson <tim+github@gfxmonk.net>
:platforms: Linux 2.6.13+.

.. ADMONITION:: About system requirements

    Recommended minimum kernel version: 2.6.25.

    Quote from the inotify(7) man page:

        "Inotify was merged into the 2.6.13 Linux kernel. The required library
        interfaces were added to glibc in version 2.4. (IN_DONT_FOLLOW,
        IN_MASK_ADD, and IN_ONLYDIR were only added in version 2.5.)"

    Therefore, you must ensure the system is running at least these versions
    appropriate libraries and the kernel.

.. ADMONITION:: About recursiveness, event order, and event coalescing

    Quote from the inotify(7) man page:

        If successive output inotify events produced on the inotify file
        descriptor are identical (same wd, mask, cookie, and name) then they
        are coalesced into a single event if the older event has not yet been
        read (but see BUGS).

        The events returned by reading from an inotify file descriptor form
        an ordered queue. Thus, for example, it is guaranteed that when
        renaming from one directory to another, events will be produced in
        the correct order on the inotify file descriptor.

        ...

        Inotify monitoring of directories is not recursive: to monitor
        subdirectories under a directory, additional watches must be created.

    This emitter implementation therefore automatically adds watches for
    sub-directories if running in recursive mode.

Some extremely useful articles and documentation:

.. _inotify FAQ: http://inotify.aiken.cz/?section=inotify&page=faq&lang=en
.. _intro to inotify: http://www.linuxjournal.com/article/8478

"""

from __future__ import with_statement
from watchdog.utils import platform

if platform.is_linux():
    import os
    import struct
    import threading
    import ctypes
    from ctypes import\
        c_int,\
        c_char_p,\
        c_uint32

    from pathtools.path import absolute_path

    from watchdog.utils import\
        has_attribute,\
        ctypes_find_library
    from watchdog.observers.api import\
        EventEmitter,\
        BaseObserver,\
        DEFAULT_EMITTER_TIMEOUT,\
        DEFAULT_OBSERVER_TIMEOUT
    from watchdog.events import\
        DirDeletedEvent,\
        DirModifiedEvent,\
        DirMovedEvent,\
        DirCreatedEvent,\
        FileDeletedEvent,\
        FileModifiedEvent,\
        FileMovedEvent,\
        FileCreatedEvent,\
        EVENT_TYPE_MODIFIED,\
        EVENT_TYPE_CREATED,\
        EVENT_TYPE_DELETED,\
        EVENT_TYPE_MOVED

    libc_string = ctypes_find_library('c', 'libc.so.6')
    libc = ctypes.CDLL(libc_string, use_errno=True)

    if (not has_attribute(libc, 'inotify_init') or
        not has_attribute(libc, 'inotify_add_watch') or
        not has_attribute(libc, 'inotify_rm_watch')):
        raise ImportError("Unsupported libc version found: %s" % libc_string)

    # #include <sys/inotify.h>
    # char *strerror(int errnum);
    #strerror = ctypes.CFUNCTYPE(c_char_p, c_int)(
    #    ("strerror", libc))

    # #include <sys/inotify.h>
    # int inotify_init(void);
    inotify_init = ctypes.CFUNCTYPE(c_int, use_errno=True)(
            ("inotify_init", libc))

    # #include <sys/inotify.h>
    # int inotify_init1(int flags);
    try:
        inotify_init1 = ctypes.CFUNCTYPE(c_int, c_int, use_errno=True)(
                ("inotify_init1", libc))
    except AttributeError:
        def inotify_init1(flags):
            raise AttributeError("No such symbol inotify_init1 in libc. Non-blocking inotify is only provided by Linux 2.6.27 and newer.")


    # #include <sys/inotify.h>
    # int inotify_add_watch(int fd, const char *pathname, uint32_t mask);
    inotify_add_watch =\
    ctypes.CFUNCTYPE(c_int, c_int, c_char_p, c_uint32, use_errno=True)(
            ("inotify_add_watch", libc))

    # #include <sys/inotify.h>
    # int inotify_rm_watch(int fd, uint32_t wd);
    inotify_rm_watch = ctypes.CFUNCTYPE(c_int, c_int, c_uint32, use_errno=True)(
            ("inotify_rm_watch", libc))

    class InotifyConstants(object):
        """
        Constants related to inotify.
        """
        # User-space events
        IN_ACCESS = 0x00000001     # File was accessed.
        IN_MODIFY = 0x00000002     # File was modified.
        IN_ATTRIB = 0x00000004     # Meta-data changed.
        IN_CLOSE_WRITE = 0x00000008     # Writable file was closed.
        IN_CLOSE_NOWRITE = 0x00000010     # Unwritable file closed.
        IN_OPEN = 0x00000020     # File was opened.
        IN_MOVED_FROM = 0x00000040     # File was moved from X.
        IN_MOVED_TO = 0x00000080     # File was moved to Y.
        IN_CREATE = 0x00000100     # Subfile was created.
        IN_DELETE = 0x00000200     # Subfile was deleted.
        IN_DELETE_SELF = 0x00000400     # Self was deleted.
        IN_MOVE_SELF = 0x00000800     # Self was moved.

        # Helper user-space events.
        IN_CLOSE = IN_CLOSE_WRITE | IN_CLOSE_NOWRITE  # Close.
        IN_MOVE = IN_MOVED_FROM | IN_MOVED_TO  # Moves.

        # Events sent by the kernel to a watch.
        IN_UNMOUNT = 0x00002000     # Backing file system was unmounted.
        IN_Q_OVERFLOW = 0x00004000     # Event queued overflowed.
        IN_IGNORED = 0x00008000     # File was ignored.

        # Special flags.
        IN_ONLYDIR = 0x01000000     # Only watch the path if it's a directory.
        IN_DONT_FOLLOW = 0x02000000     # Do not follow a symbolic link.
        IN_EXCL_UNLINK = 0x04000000     # Exclude events on unlinked objects
        IN_MASK_ADD = 0x20000000     # Add to the mask of an existing watch.
        IN_ISDIR = 0x40000000     # Event occurred against directory.
        IN_ONESHOT = 0x80000000     # Only send event once.

        # All user-space events.
        IN_ALL_EVENTS = reduce(lambda x, y: x | y, [
                IN_ACCESS,
                IN_MODIFY,
                IN_ATTRIB,
                IN_CLOSE_WRITE,
                IN_CLOSE_NOWRITE,
                IN_OPEN,
                IN_MOVED_FROM,
                IN_MOVED_TO,
                IN_DELETE,
                IN_CREATE,
                IN_DELETE_SELF,
                IN_MOVE_SELF,
                ])

        # Flags for ``inotify_init1``
        IN_CLOEXEC = 0x02000000
        IN_NONBLOCK = 0x00004000

        # All inotify bits.
        ALL_INOTIFY_BITS = reduce(lambda x, y: x | y, [
                IN_ACCESS,
                IN_MODIFY,
                IN_ATTRIB,
                IN_CLOSE_WRITE,
                IN_CLOSE_NOWRITE,
                IN_OPEN,
                IN_MOVED_FROM,
                IN_MOVED_TO,
                IN_CREATE,
                IN_DELETE,
                IN_DELETE_SELF,
                IN_MOVE_SELF,
                IN_UNMOUNT,
                IN_Q_OVERFLOW,
                IN_IGNORED,
                IN_ONLYDIR,
                IN_DONT_FOLLOW,
                IN_EXCL_UNLINK,
                IN_MASK_ADD,
                IN_ISDIR,
                IN_ONESHOT,
                ])

    # Watchdog's API cares only about these events.

    WATCHDOG_ALL_EVENTS = reduce(lambda x, y: x | y, [
            InotifyConstants.IN_CLOSE_WRITE,
            InotifyConstants.IN_ATTRIB,
            InotifyConstants.IN_MOVED_FROM,
            InotifyConstants.IN_MOVED_TO,
            InotifyConstants.IN_CREATE,
            InotifyConstants.IN_DELETE,
            ])

    class InotifyEvent(object):
        """
        Inotify event struct wrapper.

        :param wd:
            Watch descriptor
        :param mask:
            Event mask
        :param cookie:
            Event cookie
        :param name:
            Event name.
        :param src_path:
            Event source path
        """

        def __init__(self, wd, mask, cookie, name, src_path):
            self._wd = wd
            self._mask = mask
            self._cookie = cookie
            self._name = name
            self._src_path = src_path

        @property
        def src_path(self):
            return self._src_path

        @property
        def wd(self):
            return self._wd

        @property
        def mask(self):
            return self._mask

        @property
        def cookie(self):
            return self._cookie

        @property
        def name(self):
            return self._name

        # Test event types.
        @property
        def is_modify(self):
            return self._mask & InotifyConstants.IN_MODIFY > 0

        @property
        def is_close_write(self):
            return self._mask & InotifyConstants.IN_CLOSE_WRITE > 0

        @property
        def is_close_nowrite(self):
            return self._mask & InotifyConstants.IN_CLOSE_NOWRITE > 0

        @property
        def is_access(self):
            return self._mask & InotifyConstants.IN_ACCESS > 0

        @property
        def is_delete(self):
            return self._mask & InotifyConstants.IN_DELETE > 0

        @property
        def is_create(self):
            return self._mask & InotifyConstants.IN_CREATE > 0

        @property
        def is_moved_from(self):
            return self._mask & InotifyConstants.IN_MOVED_FROM > 0

        @property
        def is_moved_to(self):
            return self._mask & InotifyConstants.IN_MOVED_TO > 0

        @property
        def is_move(self):
            return self._mask & InotifyConstants.IN_MOVE > 0

        @property
        def is_attrib(self):
            return self._mask & InotifyConstants.IN_ATTRIB > 0

        # Additional bit masks
        @property
        def is_ignored(self):
            return self._mask & InotifyConstants.IN_IGNORED > 0

        @property
        def is_directory(self):
            return self._mask & InotifyConstants.IN_ISDIR > 0

        # Python-specific functionality.
        @property
        def key(self):
            return (self._src_path,
                    self._wd,
                    self._mask,
                    self._cookie,
                    self._name)

        def __eq__(self, inotify_event):
            return self.key == inotify_event.key

        def __ne__(self, inotify_event):
            return self.key == inotify_event.key

        def __hash__(self):
            return hash(self.key)


        def _get_mask_string(self, mask):
            masks = []
            for c in dir(InotifyConstants):
                if c.startswith('IN_') and c not in ['IN_ALL_EVENTS', 'IN_CLOSE'
                                                     , 'IN_MOVE']:
                    c_val = getattr(InotifyConstants, c)
                    if mask & c_val:
                        masks.append(c)
            mask_string = '|'.join(masks)
            return mask_string


        _R = "<InotifyEvent: src_path=%s, wd=%d, mask=%s, cookie=%d, name=%s>"

        def __repr__(self):
            mask_string = self._get_mask_string(self.mask)
            return InotifyEvent._R % (self.src_path,
                                      self.wd,
                                      mask_string,
                                      self.cookie,
                                      self.name)


    class inotify_event_struct(ctypes.Structure):
        """
        Structure representation of the inotify_event structure
        (used in buffer size calculations)::

            struct inotify_event {
                __s32 wd;            /* watch descriptor */
                __u32 mask;          /* watch mask */
                __u32 cookie;        /* cookie to synchronize two events */
                __u32 len;           /* length (including nulls) of name */
                char  name[0];       /* stub for possible name */
            };
        """
        _fields_ = [('wd', c_int),
                    ('mask', c_uint32),
                    ('cookie', c_uint32),
                    ('len', c_uint32),
                    ('name', c_char_p)]

    EVENT_SIZE = ctypes.sizeof(inotify_event_struct)
    DEFAULT_NUM_EVENTS = 2048
    DEFAULT_EVENT_BUFFER_SIZE = DEFAULT_NUM_EVENTS * (EVENT_SIZE + 16)

    class Inotify(object):
        """
        Linux inotify(7) API wrapper class.

        :param path:
            The directory path for which we want an inotify object.
        :param recursive:
            ``True`` if subdirectories should be monitored; ``False`` otherwise.
        :param non_blocking:
            ``True`` to initialize inotify in non-blocking mode; ``False``
            otherwise.
        """

        def __init__(self,
                     path,
                     recursive=False,
                     event_mask=WATCHDOG_ALL_EVENTS,
                     non_blocking=False):
        # The file descriptor associated with the inotify instance.
            if non_blocking:
                inotify_fd = inotify_init1(InotifyConstants.IN_NONBLOCK)
            else:
                inotify_fd = inotify_init()
            if inotify_fd == -1:
                Inotify._raise_error()
            self._inotify_fd = inotify_fd
            self._lock = threading.Lock()

            # Stores the watch descriptor for a given path.
            self._wd_for_path = dict()
            self._path_for_wd = dict()

            path = absolute_path(path)
            self._path = path
            self._event_mask = event_mask
            self._is_recursive = recursive
            self._is_non_blocking = non_blocking
            self._add_dir_watch(path, recursive, event_mask)
            self._moved_from_events = dict()

        @property
        def event_mask(self):
            """The event mask for this inotify instance."""
            return self._event_mask

        @property
        def path(self):
            """The path associated with the inotify instance."""
            return self._path

        @property
        def is_recursive(self):
            """Whether we are watching directories recursively."""
            return self._is_recursive

        @property
        def is_non_blocking(self):
            """Determines whether this instance of inotify is non-blocking."""
            return self._is_non_blocking

        @property
        def fd(self):
            """The file descriptor associated with the inotify instance."""
            return self._inotify_fd

        def clear_move_records(self):
            """Clear cached records of MOVED_FROM events"""
            self._moved_from_events = dict()

        def source_for_move(self, destination_event):
            """The source path corresponding to the given MOVED_TO event"""
            return self._moved_from_events[destination_event.cookie].src_path

        def remember_move_from_event(self, event):
            """Save this event as the source event for future MOVED_TO events to reference"""
            self._moved_from_events[event.cookie] = event

        def add_watch(self, path):
            """
            Adds a watch for the given path.

            :param path:
                Path to begin monitoring.
            """
            with self._lock:
                path = absolute_path(path)
                self._add_watch(path, self._event_mask)

        def remove_watch(self, path):
            """
            Removes a watch for the given path.

            :param path:
                Path string for which the watch will be removed.
            """
            with self._lock:
                path = absolute_path(path)
                self._remove_watch(path)

        def close(self):
            """
            Closes the inotify instance and removes all associated watches.
            """
            with self._lock:
                self._remove_all_watches()
                os.close(self._inotify_fd)

        def read_events(self, event_buffer_size=DEFAULT_EVENT_BUFFER_SIZE):
            """
            Reads events from inotify and yields them.
            """
            with self._lock:
                event_buffer = os.read(self._inotify_fd, event_buffer_size)
                event_list = []
                for wd, mask, cookie, name in Inotify._parse_event_buffer(
                        event_buffer):
                    wd_path = self._path_for_wd[wd]
                    src_path = absolute_path(os.path.join(wd_path, name))
                    inotify_event = InotifyEvent(wd, mask, cookie, name,
                                                 src_path)


                    if inotify_event.is_moved_from:
                        self.remember_move_from_event(inotify_event)
                    elif inotify_event.is_moved_to:
                        move_src_path = self.source_for_move(inotify_event)
                        if move_src_path in self._wd_for_path:
                            # update old path -> new path
                            moved_wd = self._wd_for_path[move_src_path]
                            del self._wd_for_path[move_src_path]
                            self._wd_for_path[inotify_event.src_path] = moved_wd
                            self._path_for_wd[moved_wd] = inotify_event.src_path
                        src_path = absolute_path(os.path.join(wd_path, name))
                        inotify_event = InotifyEvent(wd, mask, cookie, name,
                                                  src_path)

                    if inotify_event.is_ignored:
                    # Clean up book-keeping for deleted watches.
                        self._remove_watch_bookkeeping(src_path)
                        continue

                    event_list.append(inotify_event)

                    if inotify_event.is_directory:
                        if inotify_event.is_create:
                        # HACK: We need to traverse the directory path
                        # recursively and simulate events for newly
                        # created subdirectories/files. This will handle
                        # mkdir -p foobar/blah/bar; touch foobar/afile

                        # TODO: When a directory from another part of the
                        # filesystem is moved into a watched directory, this
                        # will not generate events for the directory tree.
                        # We need to coalesce IN_MOVED_TO events and those
                        # IN_MOVED_TO events which don't pair up with
                        # IN_MOVED_FROM events should be marked IN_CREATE
                        # instead relative to this directory.
                            self._add_watch(src_path, self._event_mask)

                            for root, dirnames, filenames in os.walk(src_path):
                                for dirname in dirnames:
                                    full_path = absolute_path(
                                            os.path.join(root, dirname))
                                    wd_dir = self._add_watch(full_path,
                                                             self._event_mask)
                                    event_list.append(InotifyEvent(wd_dir,
                                                                   InotifyConstants.IN_CREATE |
                                                                   InotifyConstants.IN_ISDIR
                                                                   ,
                                                                   0,
                                                                   dirname,
                                                                   full_path))
                                for filename in filenames:
                                    full_path = absolute_path(
                                            os.path.join(root, filename))
                                    wd_parent_dir = self._wd_for_path[
                                                    absolute_path(
                                                            os.path.dirname(
                                                                    full_path))]
                                    event_list.append(
                                            InotifyEvent(wd_parent_dir,
                                                         InotifyConstants.IN_CREATE
                                                         ,
                                                         0,
                                                         filename,
                                                         full_path))
            return event_list


        # Non-synchronized methods.
        def _add_dir_watch(self, path, recursive, mask):
            """
            Adds a watch (optionally recursively) for the given directory path
            to monitor events specified by the mask.

            :param path:
                Path to monitor
            :param recursive:
                ``True`` to monitor recursively.
            :param mask:
                Event bit mask.
            """
            if not os.path.isdir(path):
                raise OSError('Path is not a directory')
            self._add_watch(path, mask)
            if recursive:
                for root, dirnames, _ in os.walk(path):
                    for dirname in dirnames:
                        full_path = absolute_path(os.path.join(root, dirname))
                        self._add_watch(full_path, mask)

        def _add_watch(self, path, mask):
            """
            Adds a watch for the given path to monitor events specified by the
            mask.

            :param path:
                Path to monitor
            :param mask:
                Event bit mask.
            """
            wd = inotify_add_watch(self._inotify_fd,
                                   path,
                                   mask)
            if wd == -1:
                Inotify._raise_error()
            self._wd_for_path[path] = wd
            self._path_for_wd[wd] = path
            return wd

        def _remove_all_watches(self):
            """
            Removes all watches.
            """
            for wd in self._wd_for_path.values():
                del self._path_for_wd[wd]
                if inotify_rm_watch(self._inotify_fd, wd) == -1:
                    Inotify._raise_error()

        def _remove_watch_bookkeeping(self, path):
            wd = self._wd_for_path.pop(path)
            del self._path_for_wd[wd]
            return wd

        def _remove_watch(self, path):
            """
            Removes a watch for the given path.

            :param path:
                Path to remove the watch for.
            """
            wd = self._remove_watch_bookkeeping(path)
            if inotify_rm_watch(self._inotify_fd, wd) == -1:
                Inotify._raise_error()

        @staticmethod
        def _raise_error():
            """
            Raises errors for inotify failures.
            """
            _errnum = ctypes.get_errno()
            raise OSError(os.strerror(_errnum))

        @staticmethod
        def _parse_event_buffer(event_buffer):
            """
            Parses an event buffer of ``inotify_event`` structs returned by
            inotify::

                struct inotify_event {
                    __s32 wd;            /* watch descriptor */
                    __u32 mask;          /* watch mask */
                    __u32 cookie;        /* cookie to synchronize two events */
                    __u32 len;           /* length (including nulls) of name */
                    char  name[0];       /* stub for possible name */
                };

            The ``cookie`` member of this struct is used to pair two related
            events, for example, it pairs an IN_MOVED_FROM event with an
            IN_MOVED_TO event.
            """
            i = 0
            while i + 16 < len(event_buffer):
                wd, mask, cookie, length =\
                struct.unpack_from('iIII', event_buffer, i)
                name = event_buffer[i + 16:i + 16 + length].rstrip('\0')
                i += 16 + length
                yield wd, mask, cookie, name


    ACTION_EVENT_MAP = {
        (True, EVENT_TYPE_MODIFIED): DirModifiedEvent,
        (True, EVENT_TYPE_CREATED): DirCreatedEvent,
        (True, EVENT_TYPE_DELETED): DirDeletedEvent,
        (True, EVENT_TYPE_MOVED): DirMovedEvent,
        (False, EVENT_TYPE_MODIFIED): FileModifiedEvent,
        (False, EVENT_TYPE_CREATED): FileCreatedEvent,
        (False, EVENT_TYPE_DELETED): FileDeletedEvent,
        (False, EVENT_TYPE_MOVED): FileMovedEvent,
    }

    class InotifyEmitter(EventEmitter):
        """
        inotify(7)-based event emitter.

        :param event_queue:
            The event queue to fill with events.
        :param watch:
            A watch object representing the directory to monitor.
        :type watch:
            :class:`watchdog.observers.api.ObservedWatch`
        :param timeout:
            Read events blocking timeout (in seconds).
        :type timeout:
            ``float``
        """

        def __init__(self, event_queue, watch, timeout=DEFAULT_EMITTER_TIMEOUT):
            EventEmitter.__init__(self, event_queue, watch, timeout)
            self._lock = threading.Lock()
            self._inotify = Inotify(watch.path, watch.is_recursive)

        def on_thread_exit(self):
            self._inotify.close()

        def queue_events(self, timeout):
            with self._lock:
                inotify_events = self._inotify.read_events()
                if not any([event.is_moved_from or event.is_moved_to for event in inotify_events]):
                    self._inotify.clear_move_records()
                for event in inotify_events:
                    if event.is_moved_to:
                    # TODO: Sometimes this line will bomb even when a previous
                    # moved_from event with the same cookie has fired. I have
                    # yet to figure out why this is the case, so we're
                    # temporarily swallowing the exception and the move event.
                    # This happens only during massively quick file movement
                    # for example, when you execute `git gc` in a monitored
                    # directory.
                        try:
                            src_path = self._inotify.source_for_move(event)
                            to_event = event
                            dest_path = to_event.src_path

                            klass = ACTION_EVENT_MAP[
                                    (to_event.is_directory, EVENT_TYPE_MOVED)]
                            event = klass(src_path, dest_path)
                            self.queue_event(event)
                            # Generate sub events for the directory if recursive.
                            if event.is_directory and self.watch.is_recursive:
                                for sub_event in event.sub_moved_events():
                                    self.queue_event(sub_event)
                        except KeyError:
                            pass
                    elif event.is_attrib:
                        klass = ACTION_EVENT_MAP[(event.is_directory,
                                                  EVENT_TYPE_MODIFIED)]
                        self.queue_event(klass(event.src_path))
                    elif event.is_close_write:
                        klass = ACTION_EVENT_MAP[(event.is_directory,
                                                  EVENT_TYPE_MODIFIED)]
                        self.queue_event(klass(event.src_path))
                    elif event.is_delete:
                        klass = ACTION_EVENT_MAP[(event.is_directory,
                                                  EVENT_TYPE_DELETED)]
                        self.queue_event(klass(event.src_path))
                    elif event.is_create:
                        klass = ACTION_EVENT_MAP[(event.is_directory,
                                                  EVENT_TYPE_CREATED)]
                        self.queue_event(klass(event.src_path))


    class InotifyObserver(BaseObserver):
        """
        Observer thread that schedules watching directories and dispatches
        calls to event handlers.
        """

        def __init__(self, timeout=DEFAULT_OBSERVER_TIMEOUT):
            BaseObserver.__init__(self, emitter_class=InotifyEmitter,
                                  timeout=timeout)



# -*- coding: utf-8 -*-
# kqueue.py: kqueue-based event emitter.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
:module: watchdog.observers.kqueue
:synopsis: ``kqueue(2)`` based emitter implementation.
:author: Gora Khargosh <gora.khargosh@gmail.com>
:platforms: Mac OS X and BSD with kqueue(2).

.. WARNING:: kqueue is a very heavyweight way to monitor file systems.
             Each kqueue-detected directory modification triggers
             a full directory scan. Traversing the entire directory tree
             and opening file descriptors for all files will create
             performance problems. We need to find a way to re-scan
             only those directories which report changes and do a diff
             between two sub-DirectorySnapshots perhaps.

.. ADMONITION:: About ``select.kqueue`` and Python versions

    * Python 2.5 does not ship with ``select.kqueue``
    * Python 2.6 ships with a broken ``select.kqueue`` that cannot take
      multiple events in the event list passed to ``kqueue.control``.
    * Python 2.7 ships with a working ``select.kqueue``
      implementation.

    I have backported the Python 2.7 implementation to Python 2.5 and 2.6
    in the ``select_backport`` package available on PyPI.

.. ADMONITION:: About OS X performance guidelines

    Quote from the `Mac OS X File System Performance Guidelines`_:

        "When you only want to track changes on a file or directory, be sure to
        open it using the ``O_EVTONLY`` flag. This flag prevents the file or
        directory from being marked as open or in use. This is important
        if you are tracking files on a removable volume and the user tries to
        unmount the volume. With this flag in place, the system knows it can
        dismiss the volume. If you had opened the files or directories without
        this flag, the volume would be marked as busy and would not be
        unmounted."

    ``O_EVTONLY`` is defined as ``0x8000`` in the OS X header files.
    More information here: http://www.mlsite.net/blog/?p=2312

Classes
-------
.. autoclass:: KqueueEmitter
   :members:
   :show-inheritance:

Collections and Utility Classes
-------------------------------
.. autoclass:: KeventDescriptor
   :members:
   :show-inheritance:

.. autoclass:: KeventDescriptorSet
   :members:
   :show-inheritance:

.. _Mac OS X File System Performance Guidelines: http://developer.apple.com/library/ios/#documentation/Performance/Conceptual/FileSystem/Articles/TrackingChanges.html#//apple_ref/doc/uid/20001993-CJBJFIDD

"""

from __future__ import with_statement
from watchdog.utils import \
    platform #, has_attribute

if platform.is_bsd() or platform.is_darwin():
    import threading
    import errno
    import sys
    import stat
    import os

    # See the notes for this module in the documentation above ^.
    #import select
    #if not has_attribute(select, 'kqueue') or sys.version_info < (2, 7, 0):
    if sys.version_info < (2, 7, 0):
        import select_backport as select
    else:
        import select

    from pathtools.path import absolute_path

    from watchdog.observers.api import \
        BaseObserver, \
        EventEmitter, \
        DEFAULT_OBSERVER_TIMEOUT, \
        DEFAULT_EMITTER_TIMEOUT
    from watchdog.utils.dirsnapshot import DirectorySnapshot
    from watchdog.events import \
        DirMovedEvent, \
        DirDeletedEvent, \
        DirCreatedEvent, \
        DirModifiedEvent, \
        FileMovedEvent, \
        FileDeletedEvent, \
        FileCreatedEvent, \
        FileModifiedEvent, \
        EVENT_TYPE_MOVED, \
        EVENT_TYPE_DELETED, \
        EVENT_TYPE_CREATED

    # Maximum number of events to process.
    MAX_EVENTS = 4096

    # O_EVTONLY value from the header files for OS X only.
    O_EVTONLY = 0x8000

    # Pre-calculated values for the kevent filter, flags, and fflags attributes.
    if platform.is_darwin():
        WATCHDOG_OS_OPEN_FLAGS = O_EVTONLY
    else:
        WATCHDOG_OS_OPEN_FLAGS = os.O_RDONLY | os.O_NONBLOCK
    WATCHDOG_KQ_FILTER = select.KQ_FILTER_VNODE
    WATCHDOG_KQ_EV_FLAGS = select.KQ_EV_ADD | select.KQ_EV_ENABLE | select.KQ_EV_CLEAR
    WATCHDOG_KQ_FFLAGS = \
        select.KQ_NOTE_DELETE | \
        select.KQ_NOTE_WRITE | \
        select.KQ_NOTE_EXTEND | \
        select.KQ_NOTE_ATTRIB | \
        select.KQ_NOTE_LINK | \
        select.KQ_NOTE_RENAME | \
        select.KQ_NOTE_REVOKE


    # Flag tests.
    def is_deleted(kev):
        """Determines whether the given kevent represents deletion."""
        return kev.fflags & select.KQ_NOTE_DELETE

    def is_modified(kev):
        """Determines whether the given kevent represents modification."""
        fflags = kev.fflags
        return (fflags & select.KQ_NOTE_EXTEND) or (fflags & select.KQ_NOTE_WRITE)

    def is_attrib_modified(kev):
        """Determines whether the given kevent represents attribute modification."""
        return kev.fflags & select.KQ_NOTE_ATTRIB

    def is_renamed(kev):
        """Determines whether the given kevent represents movement."""
        return kev.fflags & select.KQ_NOTE_RENAME


    class KeventDescriptorSet(object):
        """
        Thread-safe kevent descriptor collection.
        """
        def __init__(self):
            # Set of KeventDescriptor
            self._descriptors = set()

            # Descriptor for a given path.
            self._descriptor_for_path = dict()

            # Descriptor for a given fd.
            self._descriptor_for_fd = dict()

            # List of kevent objects.
            self._kevents = list()

            self._lock = threading.Lock()


        @property
        def kevents(self):
            """
            List of kevents monitored.
            """
            with self._lock:
                return self._kevents

        @property
        def paths(self):
            """
            List of paths for which kevents have been created.
            """
            with self._lock:
                return self._descriptor_for_path.keys()

        def get_for_fd(self, fd):
            """
            Given a file descriptor, returns the kevent descriptor object
            for it.

            :param fd:
                OS file descriptor.
            :type fd:
                ``int``
            :returns:
                A :class:`KeventDescriptor` object.
            """
            with self._lock:
                return self._descriptor_for_fd[fd]

        def get(self, path):
            """
            Obtains a :class:`KeventDescriptor` object for the specified path.

            :param path:
                Path for which the descriptor will be obtained.
            """
            with self._lock:
                path = absolute_path(path)
                return self._get(path)

        def __contains__(self, path):
            """
            Determines whether a :class:`KeventDescriptor has been registered
            for the specified path.

            :param path:
                Path for which the descriptor will be obtained.
            """
            with self._lock:
                path = absolute_path(path)
                return self._has_path(path)

        def add(self, path, is_directory):
            """
            Adds a :class:`KeventDescriptor` to the collection for the given
            path.

            :param path:
                The path for which a :class:`KeventDescriptor` object will be
                added.
            :param is_directory:
                ``True`` if the path refers to a directory; ``False`` otherwise.
            :type is_directory:
                ``bool``
            """
            with self._lock:
                path = absolute_path(path)
                if not self._has_path(path):
                    self._add_descriptor(KeventDescriptor(path, is_directory))

        def remove(self, path):
            """
            Removes the :class:`KeventDescriptor` object for the given path
            if it already exists.

            :param path:
                Path for which the :class:`KeventDescriptor` object will be
                removed.
            """
            with self._lock:
                path = absolute_path(path)
                if self._has_path(path):
                    self._remove_descriptor(self._get(path))

        def clear(self):
            """
            Clears the collection and closes all open descriptors.
            """
            with self._lock:
                for descriptor in self._descriptors:
                    descriptor.close()
                self._descriptors.clear()
                self._descriptor_for_fd.clear()
                self._descriptor_for_path.clear()
                self._kevents = []

        # Thread-unsafe methods. Locking is provided at a higher level.
        def _get(self, path):
            """Returns a kevent descriptor for a given path."""
            return self._descriptor_for_path[path]


        def _has_path(self, path):
            """Determines whether a :class:`KeventDescriptor` for the specified
            path exists already in the collection."""
            return path in self._descriptor_for_path


        def _add_descriptor(self, descriptor):
            """
            Adds a descriptor to the collection.

            :param descriptor:
                An instance of :class:`KeventDescriptor` to be added.
            """
            self._descriptors.add(descriptor)
            self._kevents.append(descriptor.kevent)
            self._descriptor_for_path[descriptor.path] = descriptor
            self._descriptor_for_fd[descriptor.fd] = descriptor

        def _remove_descriptor(self, descriptor):
            """
            Removes a descriptor from the collection.

            :param descriptor:
                An instance of :class:`KeventDescriptor` to be removed.
            """
            self._descriptors.remove(descriptor)
            del self._descriptor_for_fd[descriptor.fd]
            del self._descriptor_for_path[descriptor.path]
            self._kevents.remove(descriptor.kevent)
            descriptor.close()


    class KeventDescriptor(object):
        """
        A kevent descriptor convenience data structure to keep together:

            * kevent
            * directory status
            * path
            * file descriptor

        :param path:
            Path string for which a kevent descriptor will be created.
        :param is_directory:
            ``True`` if the path refers to a directory; ``False`` otherwise.
        :type is_directory:
            ``bool``
        """
        def __init__(self, path, is_directory):
            self._path = absolute_path(path)
            self._is_directory = is_directory
            self._fd = os.open(path, WATCHDOG_OS_OPEN_FLAGS)
            self._kev = select.kevent(self._fd,
                                      filter=WATCHDOG_KQ_FILTER,
                                      flags=WATCHDOG_KQ_EV_FLAGS,
                                      fflags=WATCHDOG_KQ_FFLAGS)

        @property
        def fd(self):
            """OS file descriptor for the kevent descriptor."""
            return self._fd

        @property
        def path(self):
            """The path associated with the kevent descriptor."""
            return self._path

        @property
        def kevent(self):
            """The kevent object associated with the kevent descriptor."""
            return self._kev

        @property
        def is_directory(self):
            """Determines whether the kevent descriptor refers to a directory.

            :returns:
                ``True`` or ``False``
            """
            return self._is_directory

        def close(self):
            """
            Closes the file descriptor associated with a kevent descriptor.
            """
            try:
                os.close(self.fd)
            except OSError:
                pass

        @property
        def key(self):
            return (self.path, self.is_directory)

        def __eq__(self, descriptor):
            return self.key == descriptor.key

        def __ne__(self, descriptor):
            return self.key != descriptor.key

        def __hash__(self):
            return hash(self.key)

        def __repr__(self):
            return "<KeventDescriptor: path=%s, is_directory=%s>" \
                % (self.path, self.is_directory)


    class KqueueEmitter(EventEmitter):
        """
        kqueue(2)-based event emitter.

        .. ADMONITION:: About ``kqueue(2)`` behavior and this implementation

                  ``kqueue(2)`` monitors file system events only for
                  open descriptors, which means, this emitter does a lot of
                  book-keeping behind the scenes to keep track of open
                  descriptors for every entry in the monitored directory tree.

                  This also means the number of maximum open file descriptors
                  on your system must be increased **manually**.
                  Usually, issuing a call to ``ulimit`` should suffice::

                      ulimit -n 1024

                  Ensure that you pick a number that is larger than the
                  number of files you expect to be monitored.

                  ``kqueue(2)`` does not provide enough information about the
                  following things:

                  * The destination path of a file or directory that is renamed.
                  * Creation of a file or directory within a directory; in this
                    case, ``kqueue(2)`` only indicates a modified event on the
                    parent directory.

                  Therefore, this emitter takes a snapshot of the directory
                  tree when ``kqueue(2)`` detects a change on the file system
                  to be able to determine the above information.

        :param event_queue:
            The event queue to fill with events.
        :param watch:
            A watch object representing the directory to monitor.
        :type watch:
            :class:`watchdog.observers.api.ObservedWatch`
        :param timeout:
            Read events blocking timeout (in seconds).
        :type timeout:
            ``float``
        """
        def __init__(self, event_queue, watch, timeout=DEFAULT_EMITTER_TIMEOUT):
            EventEmitter.__init__(self, event_queue, watch, timeout)

            self._kq = select.kqueue()
            self._lock = threading.RLock()

            # A collection of KeventDescriptor.
            self._descriptors = KeventDescriptorSet()

            def walker_callback(path, stat_info, self=self):
                self._register_kevent(path, stat.S_ISDIR(stat_info.st_mode))
            self._snapshot = DirectorySnapshot(watch.path,
                                               watch.is_recursive,
                                               walker_callback)


        def _register_kevent(self, path, is_directory):
            """
            Registers a kevent descriptor for the given path.

            :param path:
                Path for which a kevent descriptor will be created.
            :param is_directory:
                ``True`` if the path refers to a directory; ``False`` otherwise.
            :type is_directory:
                ``bool``
            """
            try:
                self._descriptors.add(path, is_directory)
            except OSError, e:
                if e.errno == errno.ENOENT:
                    # Probably dealing with a temporary file that was created
                    # and then quickly deleted before we could open
                    # a descriptor for it. Therefore, simply queue a sequence
                    # of created and deleted events for the path.
                    #path = absolute_path(path)
                    #if is_directory:
                    #    self.queue_event(DirCreatedEvent(path))
                    #    self.queue_event(DirDeletedEvent(path))
                    #else:
                    #    self.queue_event(FileCreatedEvent(path))
                    #    self.queue_event(FileDeletedEvent(path))

                    # TODO: We could simply ignore these files.
                    # Locked files cause the python process to die with
                    # a bus error when we handle temporary files.
                    # eg. .git/index.lock when running tig operations.
                    # I don't fully understand this at the moment.
                    pass
                else:
                    # All other errors are propagated.
                    raise


        def _unregister_kevent(self, path):
            """
            Convenience function to close the kevent descriptor for a
            specified kqueue-monitored path.

            :param path:
                Path for which the kevent descriptor will be closed.
            """
            self._descriptors.remove(path)

        def queue_event(self, event):
            """
            Handles queueing a single event object.

            :param event:
                An instance of :class:`watchdog.events.FileSystemEvent`
                or a subclass.
            """
            # Handles all the book keeping for queued events.
            # We do not need to fire moved/deleted events for all subitems in
            # a directory tree here, because this function is called by kqueue
            # for all those events anyway.
            EventEmitter.queue_event(self, event)
            if event.event_type == EVENT_TYPE_CREATED:
                self._register_kevent(event.src_path, event.is_directory)
            elif event.event_type == EVENT_TYPE_MOVED:
                self._unregister_kevent(event.src_path)
                self._register_kevent(event.dest_path, event.is_directory)
            elif event.event_type == EVENT_TYPE_DELETED:
                self._unregister_kevent(event.src_path)


        def _queue_dirs_modified(self,
                                 dirs_modified,
                                 ref_snapshot,
                                 new_snapshot):
            """
            Queues events for directory modifications by scanning the directory
            for changes.

            A scan is a comparison between two snapshots of the same directory
            taken at two different times. This also determines whether files
            or directories were created, which updated the modified timestamp
            for the directory.
            """
            if dirs_modified:
                for dir_modified in dirs_modified:
                    self.queue_event(DirModifiedEvent(dir_modified))
                diff_events = new_snapshot - ref_snapshot
                for file_created in diff_events.files_created:
                    self.queue_event(FileCreatedEvent(file_created))
                for directory_created in diff_events.dirs_created:
                    self.queue_event(DirCreatedEvent(directory_created))


        def _queue_events_except_renames_and_dir_modifications(self, event_list):
            """
            Queues events from the kevent list returned from the call to
            :meth:`select.kqueue.control`.

            .. NOTE:: Queues only the deletions, file modifications,
                      attribute modifications. The other events, namely,
                      file creation, directory modification, file rename,
                      directory rename, directory creation, etc. are
                      determined by comparing directory snapshots.
            """
            files_renamed = set()
            dirs_renamed = set()
            dirs_modified = set()

            for kev in event_list:
                descriptor = self._descriptors.get_for_fd(kev.ident)
                src_path = descriptor.path

                if is_deleted(kev):
                    if descriptor.is_directory:
                        self.queue_event(DirDeletedEvent(src_path))
                    else:
                        self.queue_event(FileDeletedEvent(src_path))
                elif is_attrib_modified(kev):
                    if descriptor.is_directory:
                        self.queue_event(DirModifiedEvent(src_path))
                    else:
                        self.queue_event(FileModifiedEvent(src_path))
                elif is_modified(kev):
                    if descriptor.is_directory:
                        # When a directory is modified, it may be due to
                        # sub-file/directory renames or new file/directory
                        # creation. We determine all this by comparing
                        # snapshots later.
                        dirs_modified.add(src_path)
                    else:
                        self.queue_event(FileModifiedEvent(src_path))
                elif is_renamed(kev):
                    # Kqueue does not specify the destination names for renames
                    # to, so we have to process these after taking a snapshot
                    # of the directory.
                    if descriptor.is_directory:
                        dirs_renamed.add(src_path)
                    else:
                        files_renamed.add(src_path)
            return files_renamed, dirs_renamed, dirs_modified


        def _queue_renamed(self,
                           src_path,
                           is_directory,
                           ref_snapshot,
                           new_snapshot):
            """
            Compares information from two directory snapshots (one taken before
            the rename operation and another taken right after) to determine the
            destination path of the file system object renamed, and adds
            appropriate events to the event queue.
            """
            try:
                ref_stat_info = ref_snapshot.stat_info(src_path)
            except KeyError:
                # Probably caught a temporary file/directory that was renamed
                # and deleted. Fires a sequence of created and deleted events
                # for the path.
                if is_directory:
                    self.queue_event(DirCreatedEvent(src_path))
                    self.queue_event(DirDeletedEvent(src_path))
                else:
                    self.queue_event(FileCreatedEvent(src_path))
                    self.queue_event(FileDeletedEvent(src_path))
                # We don't process any further and bail out assuming
                # the event represents deletion/creation instead of movement.
                return

            try:
                dest_path = absolute_path(new_snapshot.path_for_inode(ref_stat_info.st_ino))
                if is_directory:
                    event = DirMovedEvent(src_path, dest_path)
                    # TODO: Do we need to fire moved events for the items
                    # inside the directory tree? Does kqueue does this
                    # all by itself? Check this and then enable this code
                    # only if it doesn't already.
                    # A: It doesn't. So I've enabled this block.
                    if self.watch.is_recursive:
                        for sub_event in event.sub_moved_events():
                            self.queue_event(sub_event)
                    self.queue_event(event)
                else:
                    self.queue_event(FileMovedEvent(src_path, dest_path))
            except KeyError:
                # If the new snapshot does not have an inode for the
                # old path, we haven't found the new name. Therefore,
                # we mark it as deleted and remove unregister the path.
                if is_directory:
                    self.queue_event(DirDeletedEvent(src_path))
                else:
                    self.queue_event(FileDeletedEvent(src_path))


        def _read_events(self, timeout=None):
            """
            Reads events from a call to the blocking
            :meth:`select.kqueue.control()` method.

            :param timeout:
                Blocking timeout for reading events.
            :type timeout:
                ``float`` (seconds)
            """
            return self._kq.control(self._descriptors.kevents,
                                    MAX_EVENTS,
                                    timeout)


        def queue_events(self, timeout):
            """
            Queues events by reading them from a call to the blocking
            :meth:`select.kqueue.control()` method.

            :param timeout:
                Blocking timeout for reading events.
            :type timeout:
                ``float`` (seconds)
            """
            with self._lock:
                try:
                    event_list = self._read_events(timeout)
                    files_renamed, dirs_renamed, dirs_modified = \
                        self._queue_events_except_renames_and_dir_modifications(event_list)

                    # Take a fresh snapshot of the directory and update the
                    # saved snapshot.
                    new_snapshot = DirectorySnapshot(self.watch.path,
                                                     self.watch.is_recursive)
                    ref_snapshot = self._snapshot
                    self._snapshot = new_snapshot

                    if files_renamed or dirs_renamed or dirs_modified:
                        for src_path in files_renamed:
                            self._queue_renamed(src_path,
                                                False,
                                                ref_snapshot,
                                                new_snapshot)
                        for src_path in dirs_renamed:
                            self._queue_renamed(src_path,
                                                True,
                                                ref_snapshot,
                                                new_snapshot)
                        self._queue_dirs_modified(dirs_modified,
                                                  ref_snapshot,
                                                  new_snapshot)
                except OSError, e:
                    if e.errno == errno.EBADF:
                        #logging.debug(e)
                        pass
                    else:
                        raise

        def on_thread_exit(self):
            # Clean up.
            with self._lock:
                self._descriptors.clear()
                self._kq.close()


    class KqueueObserver(BaseObserver):
        """
        Observer thread that schedules watching directories and dispatches
        calls to event handlers.
        """
        def __init__(self, timeout=DEFAULT_OBSERVER_TIMEOUT):
            BaseObserver.__init__(self, emitter_class=KqueueEmitter, timeout=timeout)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# watchmedo.py: Shell file monitoring utilities.
#
# Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
:module: watchdog.watchmedo
:author: Gora Khargosh <gora.khargosh@gmail.com>
:synopsis: ``watchmedo`` shell script utility.
"""

import os.path
import sys
import yaml
import time
import logging

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from argh import arg, alias, ArghParser

from watchdog.version import VERSION_STRING
from watchdog.utils import\
    read_text_file,\
    load_class
from pathtools.path import absolute_path, parent_dir_path


logging.basicConfig(level=logging.DEBUG)

CURRENT_DIR = absolute_path(os.getcwd())
DEFAULT_TRICKS_FILE_NAME = 'tricks.yaml'
DEFAULT_TRICKS_FILE_PATH = os.path.join(CURRENT_DIR, DEFAULT_TRICKS_FILE_NAME)
CONFIG_KEY_TRICKS = 'tricks'
CONFIG_KEY_PYTHON_PATH = 'python-path'


def path_split(pathname_spec, separator=os.path.sep):
    """
    Splits a pathname specification separated by an OS-dependent separator.

    :param pathname_spec:
        The pathname specification.
    :param separator:
        (OS Dependent) `:` on Unix and `;` on Windows or user-specified.
    """
    return list(pathname_spec.split(separator))


def add_to_sys_path(pathnames, index=0):
    """
    Adds specified paths at specified index into the sys.path list.

    :param paths:
        A list of paths to add to the sys.path
    :param index:
        (Default 0) The index in the sys.path list where the paths will be
        added.
    """
    for pathname in pathnames[::-1]:
        sys.path.insert(index, pathname)


def load_config(tricks_file_pathname):
    """
    Loads the YAML configuration from the specified file.

    :param tricks_file_path:
        The path to the tricks configuration file.
    :returns:
        A dictionary of configuration information.
    """
    content = read_text_file(tricks_file_pathname)
    config = yaml.load(content)
    return config


def parse_patterns(patterns_spec, ignore_patterns_spec, separator=';'):
    """
    Parses pattern argument specs and returns a two-tuple of
    (patterns, ignore_patterns).
    """
    patterns = patterns_spec.split(separator)
    ignore_patterns = ignore_patterns_spec.split(separator)
    if ignore_patterns == ['']:
        ignore_patterns = []
    return (patterns, ignore_patterns)


def observe_with(observer, event_handler, pathnames, recursive):
    """
    Single observer thread with a scheduled path and event handler.

    :param observer:
        The observer thread.
    :param event_handler:
        Event handler which will be called in response to file system events.
    :param pathnames:
        A list of pathnames to monitor.
    :param recursive:
        ``True`` if recursive; ``False`` otherwise.
    """
    for pathname in set(pathnames):
        observer.schedule(event_handler, pathname, recursive)
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


def schedule_tricks(observer, tricks, pathname, recursive):
    """
    Schedules tricks with the specified observer and for the given watch
    path.

    :param observer:
        The observer thread into which to schedule the trick and watch.
    :param tricks:
        A list of tricks.
    :param pathname:
        A path name which should be watched.
    :param recursive:
        ``True`` if recursive; ``False`` otherwise.
    """
    for trick in tricks:
        for name, value in trick.items():
            TrickClass = load_class(name)
            handler = TrickClass(**value)
            observer.schedule(handler, pathname, recursive)


@alias('tricks')
@arg('files',
     nargs='*',
     help='perform tricks from given file')
@arg('--python-path',
     default='.',
     help='paths separated by %s to add to the python path' % os.path.sep)
@arg('--interval',
     '--timeout',
     dest='timeout',
     default=1.0,
     help='use this as the polling interval/blocking timeout')
@arg('--recursive',
     default=True,
     help='recursively monitor paths')
def tricks_from(args):
    """
    Subcommand to execute tricks from a tricks configuration file.

    :param args:
        Command line argument options.
    """
    from watchdog.observers import Observer

    add_to_sys_path(path_split(args.python_path))
    observers = []
    for tricks_file in args.files:
        observer = Observer(timeout=args.timeout)

        if not os.path.exists(tricks_file):
            raise IOError("cannot find tricks file: %s" % tricks_file)

        config = load_config(tricks_file)

        try:
            tricks = config[CONFIG_KEY_TRICKS]
        except KeyError:
            raise KeyError("No `%s' key specified in %s." % (
                            CONFIG_KEY_TRICKS, tricks_file))

        if CONFIG_KEY_PYTHON_PATH in config:
            add_to_sys_path(config[CONFIG_KEY_PYTHON_PATH])

        dir_path = parent_dir_path(tricks_file)
        schedule_tricks(observer, tricks, dir_path, args.recursive)
        observer.start()
        observers.append(observer)

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        for o in observers:
            o.unschedule()
            o.stop()
    for o in observers:
        o.join()


@alias('generate-tricks-yaml')
@arg('trick_paths',
     nargs='*',
     help='Dotted paths for all the tricks you want to generate')
@arg('--python-path',
     default='.',
     help='paths separated by %s to add to the python path' % os.path.sep)
@arg('--append-to-file',
     default=None,
     help='appends the generated tricks YAML to a file; \
if not specified, prints to standard output')
@arg('-a',
     '--append-only',
     dest='append_only',
     default=False,
     help='if --append-to-file is not specified, produces output for \
appending instead of a complete tricks yaml file.')
def tricks_generate_yaml(args):
    """
    Subcommand to generate Yaml configuration for tricks named on the command
    line.

    :param args:
        Command line argument options.
    """
    python_paths = path_split(args.python_path)
    add_to_sys_path(python_paths)
    output = StringIO()

    for trick_path in args.trick_paths:
        TrickClass = load_class(trick_path)
        output.write(TrickClass.generate_yaml())

    content = output.getvalue()
    output.close()

    header = yaml.dump({CONFIG_KEY_PYTHON_PATH: python_paths})
    header += "%s:\n" % CONFIG_KEY_TRICKS
    if args.append_to_file is None:
        # Output to standard output.
        if not args.append_only:
            content = header + content
        sys.stdout.write(content)
    else:
        if not os.path.exists(args.append_to_file):
            content = header + content
        output = open(args.append_to_file, 'ab')
        output.write(content)
        output.close()


@arg('directories',
     nargs='*',
     default='.',
     help='directories to watch.')
@arg('-p',
     '--pattern',
     '--patterns',
     dest='patterns',
     default='*',
     help='matches event paths with these patterns (separated by ;).')
@arg('-i',
     '--ignore-pattern',
     '--ignore-patterns',
     dest='ignore_patterns',
     default='',
     help='ignores event paths with these patterns (separated by ;).')
@arg('-D',
     '--ignore-directories',
     dest='ignore_directories',
     default=False,
     help='ignores events for directories')
@arg('-R',
     '--recursive',
     dest='recursive',
     default=False,
     help='monitors the directories recursively')
@arg('--interval',
     '--timeout',
     dest='timeout',
     default=1.0,
     help='use this as the polling interval/blocking timeout')
@arg('--trace',
     default=False,
     help='dumps complete dispatching trace')
@arg('--debug-force-polling',
     default=False,
     help='[debug] forces polling')
@arg('--debug-force-kqueue',
     default=False,
     help='[debug] forces BSD kqueue(2)')
@arg('--debug-force-winapi',
     default=False,
     help='[debug] forces Windows API')
@arg('--debug-force-winapi-async',
     default=False,
     help='[debug] forces Windows API + I/O completion')
@arg('--debug-force-fsevents',
     default=False,
     help='[debug] forces Mac OS X FSEvents')
@arg('--debug-force-inotify',
     default=False,
     help='[debug] forces Linux inotify(7)')
def log(args):
    """
    Subcommand to log file system events to the console.

    :param args:
        Command line argument options.
    """
    from watchdog.utils import echo
    from watchdog.tricks import LoggerTrick

    if args.trace:
        echo.echo_class(LoggerTrick)

    patterns, ignore_patterns =\
    parse_patterns(args.patterns, args.ignore_patterns)
    handler = LoggerTrick(patterns=patterns,
                          ignore_patterns=ignore_patterns,
                          ignore_directories=args.ignore_directories)
    if args.debug_force_polling:
        from watchdog.observers.polling import PollingObserver as Observer
    elif args.debug_force_kqueue:
        from watchdog.observers.kqueue import KqueueObserver as Observer
    elif args.debug_force_winapi_async:
        from watchdog.observers.read_directory_changes_async import\
            WindowsApiAsyncObserver as Observer
    elif args.debug_force_winapi:
        from watchdog.observers.read_directory_changes import\
            WindowsApiObserver as Observer
    elif args.debug_force_inotify:
        from watchdog.observers.inotify import InotifyObserver as Observer
    elif args.debug_force_fsevents:
        from watchdog.observers.fsevents import FSEventsObserver as Observer
    else:
    # Automatically picks the most appropriate observer for the platform
    # on which it is running.
        from watchdog.observers import Observer
    observer = Observer(timeout=args.timeout)
    observe_with(observer, handler, args.directories, args.recursive)


@arg('directories',
     nargs='*',
     default='.',
     help='directories to watch')
@arg('-c',
     '--command',
     dest='command',
     default=None,
     help='''shell command executed in response to matching events.
These interpolation variables are available to your command string::

    ${watch_src_path}    - event source path;
    ${watch_dest_path}   - event destination path (for moved events);
    ${watch_event_type}  - event type;
    ${watch_object}      - ``file`` or ``directory``

Note::
    Please ensure you do not use double quotes (") to quote
    your command string. That will force your shell to
    interpolate before the command is processed by this
    subcommand.

Example option usage::

    --command='echo "${watch_src_path}"'
''')
@arg('-p',
     '--pattern',
     '--patterns',
     dest='patterns',
     default='*',
     help='matches event paths with these patterns (separated by ;).')
@arg('-i',
     '--ignore-pattern',
     '--ignore-patterns',
     dest='ignore_patterns',
     default='',
     help='ignores event paths with these patterns (separated by ;).')
@arg('-D',
     '--ignore-directories',
     dest='ignore_directories',
     default=False,
     help='ignores events for directories')
@arg('-R',
     '--recursive',
     dest='recursive',
     default=False,
     help='monitors the directories recursively')
@arg('--interval',
     '--timeout',
     dest='timeout',
     default=1.0,
     help='use this as the polling interval/blocking timeout')
def shell_command(args):
    """
    Subcommand to execute shell commands in response to file system events.

    :param args:
        Command line argument options.
    """
    from watchdog.observers import Observer
    from watchdog.tricks import ShellCommandTrick

    if not args.command:
        args.command = None

    patterns, ignore_patterns = parse_patterns(args.patterns,
                                               args.ignore_patterns)
    handler = ShellCommandTrick(shell_command=args.command,
                                patterns=patterns,
                                ignore_patterns=ignore_patterns,
                                ignore_directories=args.ignore_directories)
    observer = Observer(timeout=args.timeout)
    observe_with(observer, handler, args.directories, args.recursive)


epilog = """Copyright (C) 2010 Gora Khargosh <gora.khargosh@gmail.com>.

Licensed under the terms of the MIT license. Please see LICENSE in the
source code for more information."""

parser = ArghParser(epilog=epilog)
parser.add_commands([tricks_from,
                     tricks_generate_yaml,
                     log,
                     shell_command])
parser.add_argument('--version',
                    action='version',
                    version='%(prog)s ' + VERSION_STRING)


def main():
    """Entry-point function."""
    parser.dispatch()


if __name__ == '__main__':
    main()
